# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
# vim: set sts=0 ts=8 sw=8 tw=0 noet :
#
# Copyright (C) 2012 Jonathan Matthew  <jonathan@d14n.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

from gi.repository import Gio

# if libsecret isn't new enough, this will crash
from rb import rbconfig
Secret = None
if rbconfig.libsecret_enabled:
	# Till libsecret completely replaces gnome-keyring, we'll fall back to not
	# saving the password if libsecret can't be found. This code can be removed later.
	try:
		import gi
		gi.require_version('Secret', '1')
		from gi.repository import Secret
		# We need to be able to fetch passwords stored by libgnome-keyring, so we use
		# a schema with SECRET_SCHEMA_DONT_MATCH_NAME set.
		# See: http://developer.gnome.org/libsecret/unstable/migrating-schemas.html
		MAGNATUNE_SCHEMA = Secret.Schema.new("org.gnome.rhythmbox.plugins.magnatune",
							Secret.SchemaFlags.DONT_MATCH_NAME,
							{"rhythmbox-plugin": Secret.SchemaAttributeType.STRING})
	except ImportError:
		pass

if Secret is None:
	print ("You need to install libsecret and its introspection files to store your Magnatune password")

__instance = None


def instance():
	global __instance
	if __instance is None:
		__instance = MagnatuneAccount()
	return __instance

class MagnatuneAccount(object):
	def __init__(self):
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.magnatune")
		self.secret = None
		self.keyring_attributes = {"rhythmbox-plugin": "magnatune"}
		if Secret is None:
			print ("Account details will not be saved because libsecret was not found")
			return
		# Haha.
		self.secret_service = Secret.Service.get_sync(Secret.ServiceFlags.OPEN_SESSION, None)
		items = self.secret_service.search_sync(MAGNATUNE_SCHEMA,
							self.keyring_attributes,
							Secret.SearchFlags.LOAD_SECRETS,
							None)
		if not items:
			# The Python API doesn't seem to have a way to differentiate between errors and no results.
			print ("Couldn't find an existing keyring entry")
			return
		self.secret = items[0].get_secret().get().decode("utf-8")

	def get(self):
		if self.secret is None:
			return ('none', None, None)

		account_type = self.settings['account-type']
		try:
			(username, password) = self.secret.split("\n")
			return (account_type, username, password)
		except ValueError:
			return ('none', None, None)

	def update(self, username, password):
		secret = '\n'.join((username, password))
		if secret == self.secret:
			print ("Account details not changed")
			return
		self.secret = secret
		if Secret is None:
			print ("Account details were not saved because libsecret was not found")
			return
		result = Secret.password_store_sync(MAGNATUNE_SCHEMA,
							self.keyring_attributes,
							Secret.COLLECTION_DEFAULT,
							"Rhythmbox: Magnatune account information",
							secret, None)
		if not result:
			print ("Couldn't create keyring item!")



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2008, 2009, 2010 Edgar Luna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import re
import string
import rb
import stringmatch

min_artist_match = .5
min_song_match = .5

class DarkLyricsParser (object):
	"""Parser for Lyrics from www.darklyrics.com"""


	def __init__(self, artist, title):
		self.artist = artist
		self.title = title
		self.artist_ascii = ''
		self.titlenumber = ''

	def search(self, callback, *data):
		"""Do a request of a specific url based on artist's first letter name."""

		self.artist_ascii = ''.join(c for c in self.artist.lower() \
						    if c in string.ascii_letters)
		self.artist_ascii = self.artist_ascii.lower()
		firstcharurl = 'http://www.darklyrics.com/%s.html' % (self.artist_ascii[0])
		loader = rb.Loader()
		loader.get_url (firstcharurl, self.search_artist, callback, *data)

	def search_artist(self, artist_page, callback, *data):
		"""Search for the link to the page of artist in artists_page
		"""
		if artist_page is None:
			callback (None, *data)
			return
		artist_page = artist_page.decode('iso-8859-1')
		link_section = re.split ('tban.js', artist_page, 1)[1]
		pattern_link =  '<a href="'
		pattern_artist = '([^"]*)">*([^<]*)<'
		links = re.split (pattern_link, link_section.lower())
		links.pop(0)
		best_match = ()
		smvalue_bestmatch = 0
		for line in links:
			artist = re.findall(pattern_artist, line)
			if len(artist) == 0:
				continue
			artist_link, artist_name = artist[0]
			artist_url = 'http://www.darklyrics.com/%s' % (artist_link)
			if artist_link[:5] == 'http:':
				continue
			artist_name = artist_name.strip()
			smvalue = stringmatch.string_match (artist_name, self.artist_ascii)
			if smvalue > min_artist_match and smvalue > smvalue_bestmatch:
				best_match = (smvalue, artist_url, artist_name)
				smvalue_bestmatch = smvalue

		if not best_match:
			# Lyrics are located in external site
			callback (None, *data)
			return
		loader = rb.Loader ()
		self.artist  = best_match[2]
		loader.get_url (best_match[1], self.search_song, callback, *data)

	class SongFound (object):
		def __init__ (self, smvalue, title, number, album, artist):
			self.smvalue = smvalue
			self.title = title
			self.number = number
			self.album = album
			self.artist = artist

		def __str__(self):
			return '(' + str(self.smvalue) + '. ' + self.title + '. ' + self.album + '. ' + self.artist + ')'

	def search_song (self, songlist, callback, *data):
		"""If artist's page is found, search_song looks for the song.

		The artist page contains a list of all the albums and
		links to the songs lyrics from this.
		"""
		if songlist is None:
			callback (None, *data)
			return
		songlist = songlist.decode('iso-8859-1')
		# Search for all the <a>
		# filter for those that has the artist name string_match
		#        and for those which its content is artist string_match
		# Sort by values given from string_match
		# and get the best
		link_section = re.split('LYRICS</h1>', songlist)[1]
		link_section = link_section.lower()
		pattern_song = '<a href="../lyrics/(.*)/(.*).html#([^"]+)">(.*)</a>'
		matches = re.findall (pattern_song.lower(), link_section)
		best_match = ""
		for line in matches:
			artist, album, number, title = line
			smvalue = stringmatch.string_match (title.lower().replace(' ', '' ),
					   self.title.lower().replace(' ', ''))
			if smvalue > min_song_match:
				best_match  = self.SongFound(smvalue,
							     title,
							     number,
							     album,
							     artist)
				break
		if not best_match:
			callback (None, *data)
			return
		loader = rb.Loader ()
		url = 'http://www.darklyrics.com/lyrics/%s/%s.html' % (best_match.artist, best_match.album)
		self.title = best_match.title
		self.titlenumber = best_match.number
		loader.get_url (url, self.parse_lyrics, callback, *data)

	def parse_lyrics (self, album, callback, *data):
		"""In the album's page parse_lyrics get the lyrics of the song.

		This page contains all the lyrics for self.album, but
		this method get rides of everything that isn't the
		lyrics of self.title"""
		if album is None:
			callback (None, *data)
			return
		album = album.decode('iso-8859-1')
		titleline = '<a name="%s">%s. %s(.*?)</a>' % \
		    (self.titlenumber, self.titlenumber, re.escape(self.title.title()))
		lyricmatch = re.split (titleline, album)
		if len (lyricmatch) > 2:
			lyrics = lyricmatch[2]
			lyrics = lyrics.split('<h3>')[0]
			lyrics = lyrics.replace ('\r', "")
			lyrics = re.sub (r'<.*?>', "", lyrics)
			lyrics = lyrics.strip ("\n")
			title = "%s - %s\n\n" % (self.artist.title(), self.title.title())

			lyrics = title + str (lyrics)
			lyrics += "\n\nLyrics provided by Dark Lyrics"
			callback (lyrics, *data)
		else:
			callback (None, *data)
			return



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2006 - Ed Catmur <ed@catmur.co.uk>
# Copyright (C) 2009 - Jonathan Matthew <jonathan@d14n.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import os

from gi.repository import RB
from gi.repository import GObject, GLib, Gio

IMAGE_NAMES = ["cover", "album", "albumart", "front", ".folder", "folder"]
ITEMS_PER_NOTIFICATION = 10

def file_root (f_name):
	return os.path.splitext (f_name)[0].lower ()

def shared_prefix_length (a, b):
	l = 0
	while a[l] == b[l]:
		l = l+1
	return l

class LocalSearch:
	def __init__ (self):
		pass

	def finished(self, results):
		parent = self.file.get_parent()
		ordered = []
		key = RB.ExtDBKey.create_storage("album", self.album)
		key.add_field("artist", self.artists[0])

		# Compare lower case, without file extension
		for name in [file_root (self.file.get_basename())] + IMAGE_NAMES:
			for f_name in results:
				if file_root (f_name) == name:
					uri = parent.resolve_relative_path(f_name).get_uri()
					self.store.store_uri(key, RB.ExtDBSourceType.USER, uri)

		# look for file names containing the artist and album (case-insensitive)
		# (mostly for jamendo downloads)
		album = self.album.lower()
		for f_name in results:
			f_root = file_root (f_name).lower()
			for artist in self.artists:
				artist = artist.lower()
				if f_root.find (artist) != -1 and f_root.find (album) != -1:
					nkey = RB.ExtDBKey.create_storage("album", album)
					nkey.add_field("artist", artist)
					uri = parent.resolve_relative_path(f_name).get_uri()
					print("found album+artist match " + uri)
					self.store.store_uri(nkey, RB.ExtDBSourceType.USER, uri)

		# if that didn't work, look for the longest shared prefix
		# only accept matches longer than 2 to avoid weird false positives
		match_len = 2
		match = None
		for f_name in results:
			pl = shared_prefix_length(f_name, self.file.get_basename())
			if pl > match_len:
				match_len = pl
				match = f_name

		if match is not None:
			uri = parent.resolve_relative_path(match).get_uri()
			print("found prefix match " + uri)
			self.store.store_uri(key, RB.ExtDBSourceType.USER, uri)

		self.callback(self.callback_args)

	def _close_enum_cb(self, fileenum, result, results):
		try:
			fileenum.close_finish(result)
		except Exception as e:
			print("couldn't close file enumerator: %s" % e)
		

	def _enum_dir_cb(self, fileenum, result, results):
		try:
			files = fileenum.next_files_finish(result)
			if files is None or len(files) == 0:
				print("okay, done; got %d files" % len(results))
				fileenum.close_async(GLib.PRIORITY_DEFAULT, None, self._close_enum_cb, None)
				self.finished(results)
				return

			for f in files:
				ct = f.get_attribute_string("standard::content-type")
				# assume readable unless told otherwise
				readable = True
				if f.has_attribute("access::can-read"):
					readable = f.get_attribute_boolean("access::can-read")
				if ct is not None and ct.startswith("image/") and readable:
					results.append(f.get_name())

			fileenum.next_files_async(ITEMS_PER_NOTIFICATION, GLib.PRIORITY_DEFAULT, None, self._enum_dir_cb, results)
		except Exception as e:
			print("okay, probably done: %s" % e)
			import sys
			sys.excepthook(*sys.exc_info())
			self.finished(results)
			fileenum.close_async(GLib.PRIORITY_DEFAULT, None, self._close_enum_cb, None)


	def _enum_children_cb(self, parent, result, data):
		try:
			enumfiles = parent.enumerate_children_finish(result)
			enumfiles.next_files_async(ITEMS_PER_NOTIFICATION, GLib.PRIORITY_DEFAULT, None, self._enum_dir_cb, [])
		except Exception as e:
			print("okay, probably done: %s" % e)
			if not isinstance(e, GLib.GError):
				import sys
				sys.excepthook(*sys.exc_info())
			self.callback(self.callback_args)


	def search (self, key, last_time, store, callback, *args):
		# ignore last_time

		location = key.get_info("location")
		if location is None:
			print("not searching, we don't have a location")
			callback(args)
			return

		self.file = Gio.file_new_for_uri(location)

		self.album = key.get_field("album")
		if self.album is None:
			print("not searching, we don't have an album")
			callback(args)
			return

		self.artists = key.get_field_values("artist")
		self.store = store
		self.callback = callback
		self.callback_args = args

		print('searching for local art for %s' % (self.file.get_uri()))
		parent = self.file.get_parent()
		enumfiles = parent.enumerate_children_async("standard::content-type,access::can-read,standard::name", 0, 0, None, self._enum_children_cb, None)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2009 - Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

#
# string matching algorithm described here: http://www.catalysoft.com/articles/StrikeAMatch.html
#

def pairs(str):
	pairs = []
	for word in str.split(' '):
		wpairs = map(lambda x: word[x:x+2], range(len(word)-1))
		pairs.extend(wpairs)

	return pairs

def string_match(a, b):
	apairs = pairs(a.lower())
	bpairs = pairs(b.lower())

	intersection = 0
	union = len(apairs) + len(bpairs)
	for p in apairs:
		try:
			bpairs.remove(p)
			intersection = intersection + 1
		except ValueError:
			pass

	return float(intersection*2) / float(union)




# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2007 Austin  <austiny@sohu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import sys
import urllib.parse
import re
import rb
from xml.dom import minidom

class WinampcnParser(object):
	def __init__(self, artist, title):
		self.artist = artist
		self.title = title
	
	def search(self, callback, *data):

		# encode search string
		title_encode = urllib.parse.quote(self.title.replace(' ', '').encode('gbk'))
		artist_encode = urllib.parse.quote(self.artist.replace(' ', '').encode('gbk'))
		url = 'http://www.winampcn.com/lyrictransfer/get.aspx?song=%s&artist=%s&lsong=%s&Datetime=20060601' % (title_encode, artist_encode, title_encode)
		
		loader = rb.Loader()
		loader.get_url (url, self.got_lyrics, callback, *data)
		
	def got_lyrics(self, xmltext, callback, *data):
		# retrieve xml content
		if xmltext is None:
			print("no response")
			callback (None, *data)
			return
		xmltext = xmltext.decode('gbk')

		try:
			xmltext = xmltext.replace('encoding="gb2312"', 'encoding="UTF-8"')
			xmldoc = minidom.parseString(xmltext)
			root = xmldoc.documentElement

			lrcurl = root.getElementsByTagName('LyricUrl')[0].childNodes[0].data
			if lrcurl is None:
				print("no lyric urls")
				callback (xmltext, *data)
				return

			# download the lyrics file
			lrcurl = lrcurl.replace('%3A', ':');
			print("url: %s" % lrcurl)

			loader = rb.Loader()
			loader.get_url (lrcurl, self.parse_lyrics, callback, *data)
		except:
			callback (None, *data)


	def parse_lyrics(self, lyrics, callback, *data):

		if lyrics is None:
			print("no lyrics")
			callback (None, *data)
			return

		# transform it into plain text
		lrcplaintext = lyrics.decode('gbk')
		try:
			lrcplaintext = re.sub(r'\[.*?\]', '', lrcplaintext)
		except:
			print("unable to decode lyrics")
			callback (None, *data)
			return

		lrcplaintext += "\n\nLyrics provided by winampcn.com"
		callback(lrcplaintext, *data)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2006 Jonathan Matthew
# Copyright (C) 2007 James Livingston
# Copyright (C) 2007 Sirio Bolaños Puchet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import os, re
import urllib.request

import rb
from gi.repository import Gtk, Gio, GObject, Peas
from gi.repository import RB
from gi.repository import Gst, GstPbutils

import LyricsParse
from LyricsConfigureDialog import LyricsConfigureDialog

import gettext
gettext.install('rhythmbox', RB.locale_dir())


LYRIC_TITLE_STRIP=["(live[^)]*)", "(acoustic[^)]*)", "([^)]*mix)", "([^)]*version)", "([^)]*edit)", "(feat[^)]*)"]
LYRIC_TITLE_REPLACE=[("/", "-"), (" & ", " and ")]
LYRIC_ARTIST_REPLACE=[("/", "-"), (" & ", " and ")]
STREAM_SONG_TITLE='rb:stream-song-title'

def create_lyrics_view():
	tview = Gtk.TextView()
	tview.set_wrap_mode(Gtk.WrapMode.WORD)
	tview.set_editable(False)
	tview.set_left_margin(6)

	tview.set_size_request (0, 0)
	sw = Gtk.ScrolledWindow()
	sw.add(tview)
	sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
	sw.set_shadow_type(Gtk.ShadowType.IN)

	vbox = Gtk.VBox(spacing=12)
	vbox.pack_start(sw, True, True, 0)
	
	return (vbox, tview.get_buffer(), tview)

def parse_song_data(db, entry):
	(artist, title) = get_artist_and_title(db, entry)

	# don't search for 'unknown' when we don't have the artist or title information
	if artist == _("Unknown"):
		artist = ""
	if title == _("Unknown"):
		title = ""

	# convert to lowercase
	artist = artist.lower()
	title = title.lower()
	
	# replace ampersands and the like
	for exp in LYRIC_ARTIST_REPLACE:
		artist = re.sub(exp[0], exp[1], artist)
	for exp in LYRIC_TITLE_REPLACE:
		title = re.sub(exp[0], exp[1], title)

	# strip things like "(live at Somewhere)", "(accoustic)", etc
	for exp in LYRIC_TITLE_STRIP:
		title = re.sub (exp, '', title)

	# compress spaces
	title = title.strip()
	artist = artist.strip()	
	
	return (artist, title)

def get_artist_and_title(db, entry):
	stream_song_title = db.entry_request_extra_metadata(entry, STREAM_SONG_TITLE)
	if stream_song_title is not None:
		(artist, title) = extract_artist_and_title(stream_song_title)
	else:
		artist = entry.get_string(RB.RhythmDBPropType.ARTIST)
		title = entry.get_string(RB.RhythmDBPropType.TITLE)
	return (artist, title)

def extract_artist_and_title(stream_song_title):
	details = stream_song_title.split('-')
	if len(details) > 1:
		artist = details[0].strip()
		title = details[1].strip()
	else:
		details = stream_song_title.split('(')
		if len(details) > 1:
			title = details[0].strip()
			artist = details[1].strip(') ')
		else:
			title = stream_song_title
			artist = ""
	return (artist, title)
	
def build_cache_path(artist, title):
	settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.lyrics")
	folder = settings['folder']
	if folder is None or folder == "":
		folder = os.path.join(RB.user_cache_dir(), "lyrics")

	lyrics_folder = os.path.expanduser (folder)
	if not os.path.exists (lyrics_folder):
		os.mkdir (lyrics_folder)

	artist_folder = os.path.join(lyrics_folder, artist[:128])
	if not os.path.exists (artist_folder):
		os.mkdir (artist_folder)

	return os.path.join(artist_folder, title[:128] + '.lyric')

class LyricGrabber(object):
	"""
	Fetch lyrics from several sources.

	1. Local cache file
	2. Lyric tags in file meta data
	3. Online services
	"""
	def __init__(self, db, entry):
		self.db = db
		self.entry = entry

		(self.artist, self.title) = parse_song_data(self.db, self.entry)

		self.cache_path = build_cache_path(self.artist, self.title)

	def verify_lyric(self):
		return os.path.exists(self.cache_path)

	def search_lyrics(self, callback, cache_only=False):
		"""
		Fetch lyrics from cache.

		If no cache file exist, tag extraction is tried next.
		"""
		self.callback = callback

		status = self.verify_lyric()
		if status:
			f = open(self.cache_path, 'rt')
			text = f.read()
			f.close()
			self.callback(text)
		elif cache_only:
			self.callback(_("No lyrics found"))
		else:
			self.search_tags()

	def search_tags(self):
		"""
		Initiate fetching meta tags.

		Result will be handled in search_tags_result
		"""
		location = self.entry.get_playback_uri()
		self.discoverer = GstPbutils.Discoverer(timeout=Gst.SECOND*3)
		self.discoverer.connect('discovered', self.search_tags_result)
		self.discoverer.start()
		self.discoverer.discover_uri_async(location)

	def search_tags_result(self, discoverer, info, error):
		"""
		Extract lyrics from the file meta data (tags).

		If no lyrics tags are found, online services are tried next.

		Supported file formats and lyrics tags:
		- ogg/vorbis files with "LYRICS" and "SYNCLYRICS" tag
		"""
		tags = info.get_tags()
		if tags is None:
			self.search_online()
			return

		for i in range(tags.get_tag_size("extended-comment")):
			(exists, value) = tags.get_string_index("extended-comment", i)
			#ogg/vorbis unsynchronized lyrics
			if exists and value.startswith("LYRICS"):
				text = value.replace("LYRICS=", "")
				self.lyrics_found(text)
				return
			#ogg/vorbis synchronized lyrics
			elif exists and value.startswith("SYNCLYRICS"):
				text = value.replace("SYNCLYRICS=", "")
				self.lyrics_found(text)
				return

		self.search_online()

	def search_online(self):
		"""Initiate searching the online lyrics services"""
		if self.artist == "" and self.title == "":
			self.callback(_("No lyrics found"))
		else:
			parser = LyricsParse.Parser(self.artist, self.title)
			parser.get_lyrics(self.search_online_result)

	def search_online_result(self, text):
		"""Handle the result of searching online lyrics services"""
		if text is not None:
			self.lyrics_found(text)
		else:
			self.callback(_("No lyrics found"))


	def lyrics_found(self, text):
		f = open(self.cache_path, 'wt')
		f.write(text)
		f.close()

		self.callback(text)


class LyricPane(object):
	def __init__(self, db, song_info):
		self.db = db
		self.song_info = song_info
		self.entry = self.song_info.props.current_entry
		
		self.build_path()
		
		def save_lyrics(cache_path, text):
			f = open(cache_path, 'wt')
			f.write(text)
			f.close()
		
		def erase_lyrics(cache_path):
			f = open(cache_path, 'w')
			f.write("")
			f.close()
		
		def save_callback():
			buf = self.buffer
			startiter = buf.get_start_iter()
			enditer = buf.get_end_iter()
			text = buf.get_text(startiter, enditer, True)
			save_lyrics(self.cache_path, text)
			self.get_lyrics()
		
		def edit_callback(widget):
			if self.edit.get_active() == 1:
				self.tview.set_editable(True)
				self.edit.set_label(_("_Save"))
			else:
				if self.cache_path is not None:
					save_callback()
				self.tview.set_editable(False)
				self.edit.set_label(_("_Edit"))

		def discard_callback(widget):
			if self.cache_path is not None and os.path.exists(self.cache_path):
				os.remove(self.cache_path)
			self.get_lyrics()
		
		def clear_callback(widget):
			if self.cache_path is not None and os.path.exists (self.cache_path):
				erase_lyrics(self.cache_path)
			self.get_lyrics()
	   

		self.edit = Gtk.ToggleButton(label=_("_Edit"), use_underline=True)
		self.edit.connect('toggled', edit_callback)
		self.discard = Gtk.Button(label=_("_Search again"), use_underline=True)
		self.discard.connect('clicked', discard_callback)
		self.clear = Gtk.Button.new_from_stock(Gtk.STOCK_CLEAR)
		self.clear.connect('clicked', clear_callback)
		self.hbox = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
		self.hbox.set_spacing (6)
		self.hbox.set_layout(Gtk.ButtonBoxStyle.END)
		self.hbox.add(self.edit)
		self.hbox.add(self.clear)
		self.hbox.add(self.discard)
		self.hbox.set_child_secondary (self.clear, True)

		(self.view, self.buffer, self.tview) = create_lyrics_view()

		self.view.pack_start(self.hbox, False, False, 0)
		self.view.set_spacing(6)
		self.view.props.margin = 6
	
		self.view.show_all()
		self.page_num = song_info.append_page(_("Lyrics"), self.view)
		self.have_lyrics = 0
		self.visible = 0

		self.entry_change_id = song_info.connect('notify::current-entry', self.entry_changed)
		nb = self.view.get_parent()
		self.switch_page_id = nb.connect('switch-page', self.switch_page_cb)
		
		#self.get_lyrics()

	def build_path(self):
		(artist, title) = parse_song_data(self.db, self.entry)
		cache_path = build_cache_path(artist, title)
		self.cache_path = cache_path

	def entry_changed(self, pspec, duh):
		self.entry = self.song_info.props.current_entry
		self.have_lyrics = 0
		if self.visible != 0:
			self.build_path()
			self.get_lyrics()

	def switch_page_cb(self, notebook, page, page_num):
		if self.have_lyrics != 0:
			return

		if page_num != self.page_num:
			self.visible = 0
			return

		self.visible = 1
		self.get_lyrics()

	def __got_lyrics(self, text):
		self.buffer.set_text(text, -1)

	def get_lyrics(self):
		if self.entry is None:
			return

		self.buffer.set_text(_("Searching for lyrics..."), -1);
		lyrics_grabber = LyricGrabber(self.db, self.entry)
		lyrics_grabber.search_lyrics(self.__got_lyrics)


class LyricWindow (Gtk.Window):

	def __init__(self, shell):
		Gtk.Window.__init__(self)
		self.shell = shell
		self.set_border_width(12)

		close = Gtk.Button.new_from_stock(Gtk.STOCK_CLOSE)
		close.connect('clicked', lambda w: self.destroy())
	
		(lyrics_view, buffer, tview) = create_lyrics_view()
		self.buffer = buffer
		bbox = Gtk.HButtonBox()
		bbox.set_layout(Gtk.ButtonBoxStyle.END)
		bbox.pack_start(close, True, True, 0)
		lyrics_view.pack_start(bbox, False, False, 0)

		sp = shell.props.shell_player
		self.ppc_id = sp.connect('playing-song-property-changed', self.playing_property_changed)
	
		self.add(lyrics_view)
		self.set_default_size(400, 300)
		self.show_all()

	def destroy(self):
		sp = self.shell.props.shell_player
		sp.disconnect (self.ppc_id)
		Gtk.Window.destroy(self)

	def playing_property_changed(self, player, uri, prop, old_val, new_val):
		if (prop == STREAM_SONG_TITLE):
			self.update_song_lyrics(player.get_playing_entry())

	
	def __got_lyrics(self, text):
		self.buffer.set_text (text, -1)

	def update_song_lyrics(self, entry):
		db = self.shell.props.db
		(artist, title) = get_artist_and_title(db, entry)
		self.set_title(title + " - " + artist + " - " + _("Lyrics"))
		lyrics_grabber = LyricGrabber(db, entry)
		lyrics_grabber.search_lyrics(self.__got_lyrics)


class LyricsDisplayPlugin(GObject.Object, Peas.Activatable):
	__gtype_name__ = 'LyricsDisplayPlugin'
	object = GObject.property(type=GObject.Object)

	def __init__ (self):
		GObject.Object.__init__ (self)
		self.window = None

	def do_activate (self):
		shell = self.object

		self.action = Gio.SimpleAction.new("view-lyrics", None)
		self.action.connect("activate", self.show_song_lyrics, shell)
		# set accelerator?
		window = shell.props.window
		window.add_action(self.action)

		app = shell.props.application
		item = Gio.MenuItem.new(label=_("Song Lyrics"), detailed_action="win.view-lyrics")
		app.add_plugin_menu_item("view", "view-lyrics", item)

		sp = shell.props.shell_player
		self.pec_id = sp.connect('playing-song-changed', self.playing_entry_changed)
		self.playing_entry_changed (sp, sp.get_playing_entry ())

		self.csi_id = shell.connect('create_song_info', self.create_song_info)

		db = shell.props.db
		self.lyric_req_id = db.connect_after ('entry-extra-metadata-request::rb:lyrics', self.lyrics_request)

	def do_deactivate (self):
		shell = self.object
			
		app = shell.props.application
		app.remove_plugin_menu_item("view", "view-lyrics")
		app.remove_action("view-lyrics")
		self.action = None

		sp = shell.props.shell_player
		sp.disconnect (self.pec_id)

		shell.disconnect (self.csi_id)

		shell.props.db.disconnect (self.lyric_req_id)

		if self.window is not None:
			self.window.destroy ()
			self.window = None


	def show_song_lyrics (self, action, parameter, shell):

		if self.window is not None:
			self.window.destroy ()
			self.window = None

		sp = shell.props.shell_player
		entry = sp.get_playing_entry ()

		if entry is not None:
			self.window = LyricWindow(shell)
			self.window.connect("destroy", self.window_deleted)
			self.window.update_song_lyrics(entry)

	def playing_entry_changed (self, sp, entry):
		if entry is not None:
			self.action.set_enabled (True)
			if self.window is not None:
				self.window.update_song_lyrics(entry)
		else:
			self.action.set_enabled (False)

	def window_deleted (self, window):
		print("lyrics window destroyed")
		self.window = None
	
	def create_song_info (self, shell, song_info, is_multiple):
		if is_multiple is False:
			x = LyricPane(shell.props.db, song_info)

	def lyrics_request (self, db, entry):
		def lyrics_results(text):
			if text:
				db.emit_entry_extra_metadata_notify (entry, 'rb:lyrics', text)

		lyrics_grabber = LyricGrabber(db, entry)
		lyrics_grabber.search_lyrics(lyrics_results)




# Copyright (c) 2018-2019, 2022 Philipp Wolfer <ph.wolfer@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import logging
import os
from gi.repository import GLib
from client import Track

MAX_TRACKS_PER_IMPORT = 100
SUBMISSION_INTERVAL_SECONDS = 120

logger = logging.getLogger("listenbrainz")


class ListenBrainzQueue:

    def __init__(self, client):
        self.__client = client
        self.__queue = []

    def activate(self):
        self.__timeout_id = GLib.timeout_add_seconds(
            SUBMISSION_INTERVAL_SECONDS, self.submit_batch)

    def deactivate(self):
        GLib.source_remove(self.__timeout_id)

    def add(self, listened_at, track):
        try:
            # Try to submit immediately, and queue if it fails
            response = self.__client.listen(listened_at, track)
            if response.status in [401, 429] or response.status >= 500:
                self._append(listened_at, track)
        except Exception as e:
            logger.error("ListenBrainz exception %s: %s", type(e).__name__, e)
            self._append(listened_at, track)

    def load(self):
        cache_file = self.get_cache_file_path()
        if os.path.exists(cache_file):
            logger.debug("Loading queue from %s", cache_file)
            with open(cache_file) as f:
                self.__queue = json.load(f, object_hook=from_json)

    def save(self):
        cache_file = self.get_cache_file_path()
        cache_dir = os.path.dirname(cache_file)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        logger.debug("Saving queue to %s", cache_file)
        with open(cache_file, 'w') as f:
            json.dump(self.__queue, f, cls=QueueEncoder)

    def _append(self, listened_at, track):
        logger.debug("Queuing for later submission %s: %s", listened_at, track)
        self.__queue.append((listened_at, track))

    def submit_batch(self):
        if len(self.__queue) == 0:
            return True
        logger.debug("Submitting %d queued entries", len(self.__queue))
        try:
            tracks = self.__queue[0:MAX_TRACKS_PER_IMPORT]
            response = self.__client.import_tracks(tracks)
            if response.status != 200:
                return True
            if len(self.__queue) > MAX_TRACKS_PER_IMPORT:
                self.__queue = self.__queue[MAX_TRACKS_PER_IMPORT:]
            else:
                self.__queue = []
        except Exception as e:
            logger.error("ListenBrainz exception %s: %s", type(e).__name__, e)
        return True

    def get_cache_file_path(self):
        return os.path.join(GLib.get_user_cache_dir(), "rhythmbox",
                            "listenbrainz-queue.json")


class QueueEncoder(json.JSONEncoder):
    def default(self, o):
        if type(o) is Track:
            return o.to_dict()
        return super(json.JSONEncoder, self).default(o)


def from_json(json_object):
    if 'artist_name' in json_object:
        return Track.from_dict(json_object)
    return json_object



# coding: utf-8
# vim: set et sw=2:
# 
# Copyright (C) 2007-2008 - Vincent Untz
# Copyright (C) 2012 - Nirbheek Chauhan <nirbheek@gentoo.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import rb
import gi
from gi.repository import Gio, GLib, GObject, Peas
from gi.repository import RB

import gettext
gettext.install('rhythmbox', RB.locale_dir())

NORMAL_SONG_ARTIST = 'artist'
NORMAL_SONG_TITLE  = 'title'
NORMAL_SONG_ALBUM  = 'album'
STREAM_SONG_ARTIST = 'rb:stream-song-artist'
STREAM_SONG_TITLE  = 'rb:stream-song-title'
STREAM_SONG_ALBUM  = 'rb:stream-song-album'

PROPERTIES_IFACE_NAME = 'org.freedesktop.DBus.Properties'
MC5_BUS_NAME = 'org.freedesktop.Telepathy.MissionControl5'
MC5_AM_OBJ_PATH = '/org/freedesktop/Telepathy/AccountManager'
MC5_AM_IFACE_NAME = 'org.freedesktop.Telepathy.AccountManager'
MC5_ACCT_IFACE_NAME = 'org.freedesktop.Telepathy.Account'

PURPLE_BUS_NAME = 'im.pidgin.purple.PurpleService'
PURPLE_OBJ_PATH = '/im/pidgin/purple/PurpleObject'
PURPLE_IFACE_NAME = 'im.pidgin.purple.PurpleInterface'

class IMStatusPlugin (GObject.Object, Peas.Activatable):
  __gtype_name__ = 'IMStatusPlugin'
  object = GObject.property(type=GObject.Object)

  def __init__ (self):
    GObject.Object.__init__ (self)

  def _init_dbus_proxies(self):
    self.proxies = {}
    bus_type = Gio.BusType.SESSION
    flags = 0
    iface_info = None
    # Creating proxies doesn't do any blocking I/O, and never fails
    self.proxies["purple"] = Gio.DBusProxy.new_for_bus_sync(bus_type, flags, iface_info,
                               PURPLE_BUS_NAME, PURPLE_OBJ_PATH, PURPLE_IFACE_NAME, None)
    self.proxies["mc5_props"] = Gio.DBusProxy.new_for_bus_sync(bus_type, flags, iface_info,
                               MC5_BUS_NAME, MC5_AM_OBJ_PATH, PROPERTIES_IFACE_NAME, None)

  def do_activate (self):
    shell = self.object
    sp = shell.props.shell_player
    self.psc_id  = sp.connect ('playing-song-changed',
                               self.playing_entry_changed)
    self.pc_id   = sp.connect ('playing-changed',
                               self.playing_changed)
    self.pspc_id = sp.connect ('playing-song-property-changed',
                               self.playing_song_property_changed)

    self.current_entry = None
    self.current_artist = None
    self.current_title = None
    self.current_album = None

    self._init_dbus_proxies ()
    self.save_status ()

    if sp.get_playing ():
      self.set_entry (sp.get_playing_entry ())

  def do_deactivate (self):
    shell = self.object
    sp = shell.props.shell_player
    sp.disconnect (self.psc_id)
    sp.disconnect (self.pc_id)
    sp.disconnect (self.pspc_id)

    if self.current_entry is not None:
      self.restore_status ()

  def playing_changed (self, sp, playing):
    if playing:
      self.set_entry (sp.get_playing_entry ())
    else:
      self.current_entry = None
      self.restore_status ()

  def playing_entry_changed (self, sp, entry):
    if sp.get_playing ():
      self.set_entry (entry)

  def playing_song_property_changed (self, sp, uri, property, old, new):
    relevant = False
    if sp.get_playing () and property in (NORMAL_SONG_ARTIST, STREAM_SONG_ARTIST):
      self.current_artist = new
      relevant = True
    elif sp.get_playing () and property in (NORMAL_SONG_TITLE, STREAM_SONG_TITLE):
      self.current_title = new
      relevant = True
    elif sp.get_playing () and property in (NORMAL_SONG_ALBUM, STREAM_SONG_ALBUM):
      self.current_album = new
      relevant = True

    if relevant:
      self.set_status ()

  def set_entry (self, entry):
    if rb.entry_equal(entry, self.current_entry):
      return

    if self.current_entry == None:
      self.save_status ()
    self.current_entry = entry

    if entry is None:
      self.restore_status ()
      return

    self.set_status_from_entry ()

  def set_status_from_entry (self):
    shell = self.object
    db = shell.get_property ("db")
    self.current_artist = self.current_entry.get_string(RB.RhythmDBPropType.ARTIST)
    self.current_title = self.current_entry.get_string(RB.RhythmDBPropType.TITLE)
    self.current_album = self.current_entry.get_string(RB.RhythmDBPropType.ALBUM)

    if self.current_entry.get_entry_type().props.category == RB.RhythmDBEntryCategory.STREAM:
      if not self.current_artist:
        self.current_artist = db.entry_request_extra_metadata (self.current_entry, STREAM_SONG_ARTIST)
      if not self.current_title:
        self.current_title  = db.entry_request_extra_metadata (self.current_entry, STREAM_SONG_TITLE)
      if not self.current_album:
        self.current_album  = db.entry_request_extra_metadata (self.current_entry, STREAM_SONG_ALBUM)

    self.set_status ()

  def set_status (self):
    subs = {
        'artist': self.current_artist,
        'album': self.current_album,
        'title': self.current_title
    }
    if self.current_artist:
      if self.current_title:
        # Translators: do not translate %(artist)s or %(title)s, they are
        # string substitution markers (like %s) for the artist and title of
        # the current playing song.  They can be reordered if necessary.
        new_status = _(u"♫ %(artist)s - %(title)s ♫") % subs
      elif self.current_album:
        # Translators: do not translate %(artist)s or %(album)s, they are
        # string substitution markers (like %s) for the artist and album name
        # of the current playing song.  They can be reordered if necessary.
        new_status = _(u"♫ %(artist)s - %(album)s ♫") % subs
    elif self.current_album:
      # Translators: do not translate %(album)s, it is a string substitution
      # marker (like %s) for the album name of the current playing song.
      new_status = _(u"♫ %(album)s ♫") % subs
    elif self.current_title:
      # Translators: do not translate %(title)s, it is a string substitution
      # marker (like %s) for the title of the current playing song.
      new_status = _(u"♫ %(title)s ♫") % subs
    else:
      new_status = _(u"♫ Listening to music... ♫")

    self.set_mc5_status (new_status)
    self.set_purple_status (new_status)

  def save_status (self):
    self.saved_mc5 = self.get_mc5_status ()
    self.saved_purple = self.get_purple_status ()

  def restore_status (self):
    if self.saved_mc5 != None:
      self.set_mc5_status (self.saved_mc5)
    if self.saved_purple != None:
      self.set_purple_status (self.saved_purple)

  def set_mc5_status (self, new_status):
    try:
      proxy = self.proxies["mc5_props"]
      for acct_obj_path in proxy.Get("(ss)", MC5_AM_IFACE_NAME, "ValidAccounts"):
        # Create a new proxy connected to acct_obj_path
        acct_proxy = Gio.DBusProxy.new_for_bus_sync(Gio.BusType.SESSION, 0, None,
                                                    MC5_BUS_NAME, acct_obj_path,
                                                    PROPERTIES_IFACE_NAME, None)
        # status = (state, status, status_message)
        status = acct_proxy.Get("(ss)", MC5_ACCT_IFACE_NAME, "RequestedPresence")
        # Create the (uss) GVariant to set the new status message
        vstatus = GLib.Variant("(uss)", (status[0], status[1], new_status))
        # Set the status!
        acct_proxy.Set("(ssv)", MC5_ACCT_IFACE_NAME, "RequestedPresence", vstatus)
    except GLib.GError as e:
      print("GError while setting status: " + str(e))

  def get_mc5_status (self):
    try:
      proxy = self.proxies["mc5_props"]
      got_status = False
      # a bit awful: this just returns the status text from the first account
      # that has one.
      for acct_obj_path in proxy.Get("(ss)", MC5_AM_IFACE_NAME, "ValidAccounts"):
        # Create a new proxy connected to acct_obj_path
        acct_proxy = Gio.DBusProxy.new_for_bus_sync (Gio.BusType.SESSION, 0, None,
                                                     MC5_BUS_NAME, acct_obj_path,
                                                     PROPERTIES_IFACE_NAME, None)
        # Get (state, status, status_message)
        ret = acct_proxy.Get("(ss)", MC5_ACCT_IFACE_NAME, "RequestedPresence")
        got_status = True
        if ret[2] != "":
          return ret[2]
      # if all accounts have empty status, return that
      if got_status:
        return ""
    except GLib.GError as e:
      print("GError while setting status: " + str(e))
    return None

  def set_purple_status (self, new_status):
    try:
      proxy = self.proxies["purple"]
      status = proxy.PurpleSavedstatusGetCurrent()
      proxy.PurpleSavedstatusSetMessage("(is)", status, new_status)
      proxy.PurpleSavedstatusActivate("(i)", status)
    except GLib.GError as e:
      print("GError while setting status: " + str(e))

  def get_purple_status (self):
    try:
      proxy = self.proxies["purple"]
      status = proxy.PurpleSavedstatusGetCurrent()
      return proxy.PurpleSavedstatusGetMessage("(i)", status)
    except GLib.GError as e:
      print("GError while setting status: " + str(e))
    return None



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*- 
#
# Copyright (C) 2009 - Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import gi
gi.require_version('Soup', '3.0')
from gi.repository import GObject, GLib, Gio, Soup
import sys

from rbconfig import rhythmbox_version

def call_callback(callback, data, args):
	try:
		v = callback(data, *args)
		return v
	except Exception as e:
		sys.excepthook(*sys.exc_info())

loader_session = None

class Loader(object):
	def __init__ (self):
		global loader_session
		if loader_session is None:
			loader_session = Soup.Session()
			loader_session.set_user_agent("Rhythmbox/" + rhythmbox_version)
		self._cancel = Gio.Cancellable()

	def _message_cb(self, session, result, _data):
		message = session.get_async_result_message(result)
		status = message.get_status()
		body = None
		try:
			if status == 200:
				bytes = session.send_and_read_finish(result)
				if bytes:
					body = bytes.get_data()
		finally:
			call_callback(self.callback, body, self.args)

	def get_url (self, url, callback, *args):
		self.url = url
		self.callback = callback
		self.args = args
		try:
			global loader_session
			req = Soup.Message.new("GET", url)
			loader_session.send_and_read_async(
				req, GLib.PRIORITY_DEFAULT, self._cancel,
				self._message_cb, None)
		except Exception as e:
			sys.excepthook(*sys.exc_info())
			callback(None, *args)

	def cancel (self):
		self._cancel.cancel()





# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2007 James Livingston
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import rb
from gi.repository import GObject, Gio

from LyricsSites import lyrics_sites

class Parser (object):
	def __init__(self, artist, title):
		self.title = title
		self.artist = artist

		try:
			settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.lyrics")
			self.sites = settings['sites']
		except GLib.GError as e:
			print(e)
			self.sites = []

	def searcher(self, plexer, callback, *data):
		for site in lyrics_sites:
			if site['id'] not in self.sites:
				print(site['id'] + " search is disabled")
				continue

			plexer.clear()
			parser = site['class'] (self.artist, self.title)
			print("searching " + site['id'] + " for lyrics")

			parser.search(plexer.send())
			yield None

			_, (lyrics,) = plexer.receive()
			if lyrics is not None:
				callback (lyrics, *data)
				return

		callback (None, *data)

	def get_lyrics(self, callback, *data):
		rb.Coroutine (self.searcher, callback, *data).begin ()



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2009 Hardy Beltran Monasterios
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.


import urllib.parse
import rb
import re
import sys

# Deal with html entities and utf-8
# code taken from django/utils/text.py

from html.entities import name2codepoint

pattern = re.compile(r"&(#?\w+?);")

def _replace_entity(match):
	text = match.group(1)
	if text[0] == u'#':
		text = text[1:]
		try:
			if text[0] in u'xX':
				c = int(text[1:], 16)
			else:
				c = int(text)
			return chr(c)
		except ValueError:
			return match.group(0)
	else:
		try:
			return chr(name2codepoint[text])
		except (ValueError, KeyError):
			return match.group(0)

def unescape_entities(text):
	return pattern.sub(_replace_entity, text)

class TerraParser (object):
	def __init__(self, artist, title):
		self.artist = artist
		self.title = title

	def search(self, callback, *data):
		path = 'http://letras.mus.br/'

		artist = urllib.parse.quote(self.artist)
		title = urllib.parse.quote(self.title)
		join = urllib.parse.quote(' - ')

		wurl = 'winamp.php?t=%s%s%s' % (artist, join, title)
		print("search URL: " + wurl)

		loader = rb.Loader()
		loader.get_url (path + wurl, self.got_lyrics, callback, *data)

	def got_lyrics(self, result, callback, *data):
		if result is None:
			callback (None, *data)
			return

		if result is not None:
			result = result.decode('utf-8')
			if re.search('M&uacute;sica n&atilde;o encontrada', result):
				print("not found")
				callback (None, *data)
			elif re.search('<div id="letra">', result):
				callback(self.parse_lyrics(result), *data)
			else:
				callback (None, *data)
		else:
			callback (None, *data)


	def parse_lyrics(self, source):
		def unspace(x):
			return " ".join(x.split())
		def untag(x):
			return re.sub('<.*?>', '\n', x)

		source = re.split('<div id="letra">', source)[1]
		source = re.split('</?div.*?>', source)
		# source[1] = artist+title
		# source[3] = lyrics

		header = "".join(source[1].splitlines())
		# <h1><a>title</a></h1> <h2><a>artist</a></h2>
		bits = re.findall('<h.>(.*?)</h.>', header)
		artistitle = unspace(untag(" - ".join(bits)))

		lyrics = unescape_entities(artistitle) + "\n" + unescape_entities(untag(source[3]))
		lyrics += "\n\nEsta letra foi disponibilizada pelo site\nhttp://letras.mus.br"
		return lyrics



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2016 Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import gi
gi.require_version('Soup', '3.0')
from gi.repository import GLib, GObject, Gio, Peas, PeasGtk, Soup, Gtk
from gi.repository import RB
import rb

import sys
import os.path
import json
import re
import time
import struct

import siphash

import gettext
gettext.install('rhythmbox', RB.locale_dir())

def get_host_name():
        try:
                p = Gio.DBusProxy.new_for_bus_sync(Gio.BusType.SYSTEM,
                                                   0,
                                                   None,
                                                   'org.freedesktop.Avahi',
                                                   '/',
                                                   'org.freedesktop.Avahi.Server')
                return p.GetHostNameFqdn()
        except Exception as e:
                # ignore
                import socket
                return socket.gethostname()



class ClientSession(object):

	def __init__(self, plugin, connection, connid):
		print("new connection attached")
		self.connid = connid
		self.conn = connection 
		self.conn.connect("message", self.message_cb)
		self.conn.connect("closed", self.closed_cb)
		self.plugin = plugin
		self.actions = {
			'status': self.plugin.client_status,
			'next': self.plugin.client_next,
			'previous': self.plugin.client_previous,
			'playpause': self.plugin.client_playpause,
			'seek': self.plugin.client_seek
		}

	def message_cb(self, conn, msgtype, message):
		if msgtype != Soup.WebsocketDataType.TEXT:
			print("binary message received?")
			return

		d = message.get_data().decode("utf-8")
		print("message received: %s" % d)
		try:
			m = json.loads(d)
			action = m.get('action')
			print("doing %s" % action)
			if action in self.actions:
				r = self.actions[action](m)
			else:
				r = {'result': 'what'}

			print("responding %s" % str(r))
			self.conn.send_text(json.dumps(r))
		except Exception as e:
			sys.excepthook(*sys.exc_info())

	def closed_cb(self, conn):
		self.plugin.player_websocket_closed(self, self.connid)

	def dispatch(self, message):
		self.conn.send_text(message)

	def disconnect(self):
		self.conn.close(0, "")

class TrackStreamer(object):
	def __init__(self, server, message, track, content_type):
		self.server = server
		self.message = message
		self.message.connect("wrote-chunk", self.wrote_chunk)
		self.trackfile = Gio.File.new_for_uri(track)
		self.content_type = content_type
		self.stream = None
		self.offset = 0
		self.done = False
		print("streaming " + track)

	def wrote_chunk(self, msg):
		if not self.done:
			self.server.pause_message(self.message)

	def open(self):
		print("opening")
		self.trackfile.read_async(GLib.PRIORITY_DEFAULT, None, self.opened)
		self.server.pause_message(self.message)

	def opened(self, obj, result):
		try:
			print("track opened")
			headers = self.message.props.response_headers
			headers.set_content_type(self.content_type)
			headers.set_encoding(Soup.Encoding.CHUNKED)

			body = self.message.props.response_body
			body.set_accumulate(False)

			self.stream = self.trackfile.read_finish(result)
			self.message.set_status(200)
			self.read_more()
		except Exception as e:
			sys.excepthook(*sys.exc_info())
			self.message.set_status(500)
			self.server.unpause_message(self.message)

	def read_more(self):
		self.stream.read_bytes_async(65536, GLib.PRIORITY_DEFAULT, None, self.read_done)

	def read_done(self, obj, result):
		body = self.message.props.response_body
		try:
			b = self.stream.read_bytes_finish(result)
			if b.get_size() == 0:
				self.done = True
				body.complete()
			else:
				self.offset = self.offset + b.get_size()
				body.append(b.get_data()) # uh..
				self.read_more()
			self.server.unpause_message(self.message)

		except Exception as e:
			sys.excepthook(*sys.exc_info())
			if (self.offset == 0):
				self.message.set_status(500)
			else:
				body.complete()
			self.server.unpause_message(self.message)



class WebRemotePlugin(GObject.Object, Peas.Activatable):
	__gtype_name = 'WebRemotePlugin'
	object = GObject.property(type=GObject.GObject)

	signature_max_age = 60
	# we don't really need a huge replay memory.
	# clients should only make requests every couple of minutes,
	# and signatures are only valid for (currently) up to two minutes.
	replay_memory = 20

	string_props = {
		RB.RhythmDBPropType.TITLE: 'title',
		RB.RhythmDBPropType.ARTIST: 'artist',
		RB.RhythmDBPropType.ALBUM: 'album',
		RB.RhythmDBPropType.ALBUM_ARTIST: 'album-artist',
		RB.RhythmDBPropType.GENRE: 'genre',
		RB.RhythmDBPropType.COMPOSER: 'composer',
		RB.RhythmDBPropType.TITLE: 'title',
	}
	ulong_props = {
		RB.RhythmDBPropType.ENTRY_ID: 'id',
		RB.RhythmDBPropType.YEAR: 'year',
		#RB.RhythmDBPropType.BEATS_PER_MINUTE: 'bpm',
		RB.RhythmDBPropType.BITRATE: 'bitrate',
		RB.RhythmDBPropType.DURATION: 'duration',
		RB.RhythmDBPropType.TRACK_NUMBER: 'track-number',
		RB.RhythmDBPropType.TRACK_TOTAL: 'track-total',
		RB.RhythmDBPropType.DISC_NUMBER: 'disc-number',
		RB.RhythmDBPropType.DISC_TOTAL: 'disc-total'
	}

	def __init__(self):
		GObject.Object.__init__(self)
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.webremote")
		self.settings.connect("changed", self.settings_changed_cb)
		self.server = None
		self.next_connid = 0
		self.connections = {}
		self.replay = []
		self.access_key = None

		self.listen_reset = False

	def get_sign_key(self, id):
		# some day there will be multiple keys
		a = self.settings['access-key']

		ea = a.encode()
		pa = (a + 4 * '\0').encode()

		k = [0, 0, 0, 0]
		i = 0
		ki = 0
		uint = struct.Struct("<I")
		while i < len(ea):
			k[ki] = (k[ki] + uint.unpack(pa[i:i+4])[0]) % 0xffffffff
			i = i + 4
			ki = (ki + 1) % 4

		return k

	def check_http_signature(self, path, query):
		try:
			qargs = dict([b.split("=") for b in query.split("&")])
			ts = qargs['ts']
			sig = qargs['sig']
			keyid = qargs.get("k", "default")	# not used yet

			if sig in self.replay:
				print("replayed signature " + sig + " in request for " + path)
				return False

			its = int(ts) / 1000
			now = time.time()
			max = self.signature_max_age
			print("request timestamp: " + ts + ", min: " + str(now - max) + ", max: " + str(now + max))
			if (its < (now - max)) or (its > (now + max)):
				return False

			message = (path + "\n" + ts).encode()
			check = siphash.SipHash_2_4(self.get_sign_key(keyid), message).hexdigest()
			print("request signature: " + sig + ", expecting: " + check.decode())
			if check == sig.encode():
				self.replay.insert(0, sig)
				self.replay = self.replay[:self.replay_memory]
				return True

			return False

		except Exception as e:
			sys.excepthook(*sys.exc_info())
			return False

	def check_http_msg_signature(self, msg):
		u = msg.get_uri()
		return self.check_http_signature(u.get_path(), u.get_query())

	def player_websocket_cb(self, server, msg, path, conn):
		(upath, query) = path.split(":", 1)
		if self.check_http_signature(upath, query) is False:
			conn.close(403, "whatever")
			return

		cs = ClientSession(self, conn, self.next_connid)
		self.connections[self.next_connid] = cs
		self.next_connid = self.next_connid + 1

	def player_websocket_closed(self, connection, connid):
		self.connections.pop(connid)

	def dispatch(self, message):
		m = json.dumps(message)
		for c in self.connections.values():
			c.dispatch(m)

	def album_art_filename(self, filename):
		if filename is None:
			return None

		if filename.startswith(self.artcache) is False:
			return None

		rfn = filename[len(self.artcache):].lstrip('/')
		return os.path.normpath(rfn)


	def entry_details(self, entry):
		m = {}
		if entry is not None:
			for (p, k) in self.string_props.items():
				m[k] = entry.get_string(p)
			for (p, k) in self.ulong_props.items():
				m[k] = entry.get_ulong(p)

			key = entry.create_ext_db_key(RB.RhythmDBPropType.ALBUM)
			(filename, lkey) = self.art_store.lookup(key)
			m['albumart'] = self.album_art_filename(filename)
		return m

	def set_playing_position(self, update):
		try:
			(r, pos) = self.shell_player.get_playing_time()
			update['position'] = pos * 1000
		except Exception as e:
			pass

	def client_status(self, message):
		entry = self.shell_player.get_playing_entry()
		if entry:
			m = self.entry_details(entry)
			self.set_playing_position(m)
			p = self.shell_player.get_playing()
			m['playing'] = p[1]
		else:
			m = { 'playing': False, 'id': 0 }

		m['hostname'] = GLib.get_host_name()
		return m

	def client_next(self, message):
		try:
			self.shell_player.do_next()
			return {'result': 'ok'}
		except Exception as e:
			return {'result': str(e) }

	def client_previous(self, message):
		try:
			self.shell_player.do_previous()
			return {'result': 'ok'}
		except Exception as e:
			return {'result': str(e) }

	def client_playpause(self, message):
		try:
			self.shell_player.playpause()
			return {'result': 'ok'}
		except Exception as e:
			return {'result': str(e) }

	def client_seek(self, message):
		try:
			self.shell_player.set_playing_time(message['time'])
			return {'result': 'ok'}
		except Exception as e:
			return {'result': str(e) }


	def playing_song_changed_cb(self, player, entry):
		self.elapsed = 0
		self.dispatch(self.entry_details(entry))

	def playing_changed_cb(self, player, playing):
		u = { 'playing': playing }
		self.set_playing_position(u)
		self.dispatch(u)

	def playing_song_property_changed_cb(self, player, uri, prop, oldvalue, newvalue):
		if prop in self.string_props:
			self.dispatch({ self.string_props[prop]: newvalue })
		if prop in self.ulong_props:
			self.dispatch({ self.ulong_props[prop]: newvalue })

	def elapsed_nano_changed_cb(self, player, elapsed):
		if abs(elapsed - self.elapsed) > 1000000000:
			self.dispatch({'position': elapsed/1000000})	# ms
		self.elapsed = elapsed

	def art_added_cb(self, store, key, filename, data):
		entry = self.shell_player.get_playing_entry()
		if entry is not None and self.db.entry_matches_ext_db_key(entry, key):
			self.dispatch({'albumart': self.album_art_filename(filename)})

	def send_file_response(self, msg, filename, content_type):
		try:
			fp = open(filename, 'rb')
			d = fp.read()

			if callable(content_type):
				content_type = content_type(d)

			msg.set_response(content_type, Soup.MemoryUse.COPY, d)
			msg.set_status(200)
		except Exception as e:
			sys.excepthook(*sys.exc_info())
			msg.set_status(500)

	def image_content_type(self, data):
		# superhacky
		if data[1:4] == b'PNG':
			return "image/png"
		elif data[0:5] == b'<?xml':
			return "image/svg+xml"
		elif data[0:4] == b'<svg':
			return "image/svg+xml"
		else:
			return "image/jpeg"

	def http_track_cb(self, server, msg, path, query):

		if self.check_http_msg_signature(msg) is False:
			msg.set_status(403)
			return

		entry = self.shell_player.get_playing_entry()
		if entry is None:
			msg.set_status(404)
			return

		mt = entry.get_string(RB.RhythmDBPropType.MEDIA_TYPE)
		ct = RB.gst_media_type_to_mime_type(mt)
		s = TrackStreamer(server, msg, entry.get_playback_uri(), ct)
		try:
			s.open()
		except Exception as e:
			sys.excepthook(*sys.exc_info())
			msg.set_status(500)


	def http_art_cb(self, server, msg, path, query):
		if self.check_http_msg_signature(msg) is False:
			msg.set_status(403)
			return

		if msg.get_method() != "GET":
			msg.set_status(404)
			return

		if re.match("/art/[^/][a-zA-Z0-9/]+", path) is None:
			msg.set_status(404)
			return

		artpath = os.path.join(self.artcache, path[len("/art/"):])
		if not os.path.exists(artpath):
			msg.set_status(404)
			return

		self.send_file_response(msg, artpath, self.image_content_type)

	def http_icon_cb(self, server, msg, path, query):
		if msg.get_method() != "GET":
			msg.set_status(404)
			return

		if re.match("/icon/[a-zA-Z0-9.-]+/[0-9]+", path) is None:
			msg.set_status(404)
			return

		bits = path.split("/")
		iconname = bits[2]
		iconsize = int(bits[3])
		icon = Gtk.IconTheme.get_default().lookup_icon(iconname, iconsize, Gtk.IconLookupFlags.FORCE_SVG)
		if icon is None:
			msg.set_status(404)
			return

		iconfile = icon.get_filename()
		try:
			res = Gio.resources_lookup_data(iconfile, 0)
			data = res.get_data()
			content_type = self.image_content_type(data)
			msg.set_response(content_type, Soup.MemoryUse.COPY, data)
			msg.set_status(200)
		except gi.repository.GLib.GError as ge:
			# assume we couldn't find the resource, so try it as a filename
			self.send_file_response(msg, icon.get_filename(), self.image_content_type)
		except Exception as e:
			sys.excepthook(*sys.exc_info())
			msg.set_status(500)


	def serve_static(self, msg, path, subdir, content_type):

		if subdir == '':
			ssubdir = '/'
		else:
			ssubdir = "/" + subdir + "/"
		if not path.startswith(ssubdir):
			msg.set_status(403)
			return

		relpath = path[len(ssubdir):]

		if msg.get_method() != "GET" or relpath.find("/") != -1:
			msg.set_status(403)
			return

		f = rb.find_plugin_file(self, os.path.join(subdir, relpath))
		if f is None:
			msg.set_status(403)
			return

		self.send_file_response(msg, f, content_type)


	def http_static_css_cb(self, server, msg, path, query):
		self.serve_static(msg, path, "css", "text/css")

	def http_static_js_cb(self, server, msg, path, query):
		self.serve_static(msg, path, "js", "text/javascript")

	def http_root_cb(self, server, msg, path, query):
		self.serve_static(msg, '/webremote.html', '', 'text/html')

	def settings_changed_cb(self, settings, key):
		if key == 'listen-port':
			if self.http_server is not None and self.listen_reset is False:
				self.http_server.disconnect()
				self.http_listen()
		elif key == 'access-key':
			for c in self.connections.values():
				c.disconnect()

	def http_listen(self):
		print("relistening")
		port = self.settings['listen-port']
		if port != 0:
			try:
				self.http_server.listen_all(port, 0)
			except Exception as e:
				port = 0

		if port == 0:
			print("trying again")
			self.listen_reset = True
			self.http_server.listen_all(0, 0)
			# remember the port number for convenience
			uris = self.http_server.get_uris()
			if len(uris) > 0:
				self.settings['listen-port'] = uris[0].get_port()

			self.listen_reset = False


	def do_activate(self):
		shell = self.object
		self.db = shell.props.db

		self.artcache = os.path.join(RB.user_cache_dir(), "album-art", "")
		self.art_store = RB.ExtDB(name="album-art")
		self.art_store.connect("added", self.art_added_cb)

		self.shell_player = shell.props.shell_player
		self.shell_player.connect("playing-song-changed", self.playing_song_changed_cb)
		self.shell_player.connect("playing-song-property-changed", self.playing_song_property_changed_cb)
		self.shell_player.connect("playing-changed", self.playing_changed_cb)
		self.shell_player.connect("elapsed-nano-changed", self.elapsed_nano_changed_cb)
		self.playing_song_changed_cb(self.shell_player, self.shell_player.get_playing_entry())

		self.http_server = Soup.Server()
		self.http_server.add_handler(path="/art/", callback=self.http_art_cb)
		self.http_server.add_handler(path="/icon/", callback=self.http_icon_cb)
		self.http_server.add_handler(path="/entry/current/stream", callback=self.http_track_cb)
		self.http_server.add_handler(path="/css/", callback=self.http_static_css_cb)
		self.http_server.add_handler(path="/js/", callback=self.http_static_js_cb)
		self.http_server.add_websocket_handler("/ws/player", None, None, self.player_websocket_cb)
		self.http_server.add_handler(path="/", callback=self.http_root_cb)

		self.http_listen()


	def do_deactivate(self):
		self.dispatch({'shutdown': True })
		self.server = None
		self.connections = {}

class WebRemoteConfig(GObject.Object, PeasGtk.Configurable):
	__gtype_name__ = 'WebRemoteConfig'
	object = GObject.property(type=GObject.Object)

	def accesskey_focus_out_cb(self, widget, event):
		k = widget.get_text()
		if k != self.access_key:
			print("changing access key to %s" % k)
			self.access_key = k
			self.settings['access-key'] = k

		return False

	def update_port(self):
		hostname = get_host_name()
		port = self.settings['listen-port']
		url = 'http://%s:%d/' % (hostname, port)
		label = _('Launch web remote control')
		self.launch_link.set_markup('<a href="%s">%s</a>' % (url, label))

		self.portnumber.set_text("%d" % port)

	def settings_changed_cb(self, settings, key):
		if key == 'listen-port':
			self.update_port()

	def do_create_configure_widget(self):
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.webremote")
		self.settings.connect("changed", self.settings_changed_cb)

		ui_file = rb.find_plugin_file(self, "webremote-config.ui")
		self.builder = Gtk.Builder()
		self.builder.add_from_file(ui_file)

		content = self.builder.get_object("webremote-config")

		self.portnumber = self.builder.get_object("portnumber")
		self.launch_link = self.builder.get_object("launch-link")
		self.update_port()

		self.key_entry = self.builder.get_object("accesskey")
		self.access_key = self.settings['access-key']
		if self.access_key:
			self.key_entry.set_text(self.access_key)
		self.key_entry.connect("focus-out-event", self.accesskey_focus_out_cb)

		return content


GObject.type_register(WebRemoteConfig)



# -*- coding: utf-8 -*-
#
# pythonconsole.py
#
# Copyright (C) 2006 - Steve Frécinaux
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

# Parts from "Interactive Python-GTK Console" (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from gedit Python Console Plugin
#     Copyrignt (C), 2005 Raphaël Slinckx

import string
import sys
import re
import os
import traceback

from gi.repository import GLib, Gtk, Gdk, Gio, GObject, Pango, Peas
from gi.repository import RB

import gettext
gettext.install('rhythmbox', RB.locale_dir())

DEBUG_PORT = 5678

try:
	import rpdb2
	have_rpdb2 = True
except:
	have_rpdb2 = False

try:
	import debugpy
	have_debugpy = True
except:
	have_debugpy = False


class PythonConsolePlugin(GObject.Object, Peas.Activatable):
	__gtype_name__ = 'PythonConsolePlugin'

	object = GObject.property (type = GObject.Object)

	def __init__(self):
		GObject.Object.__init__(self)
		self.window = None
		
	def do_activate(self):
		shell = self.object
		app = shell.props.application

		action = Gio.SimpleAction.new("python-console", None)
		action.connect('activate', self.show_console, shell)
		app.add_action(action)

		app.add_plugin_menu_item("tools",
					 "python-console",
					 Gio.MenuItem.new(label=_("Python Console"),
						 	  detailed_action="app.python-console"))

		self.rpdb2_action = Gio.SimpleAction.new("python-debugger-winpdb", None)
		self.rpdb2_action.connect('activate', self.attach_winpdb, shell)
		app.add_action(self.rpdb2_action)

		app.add_plugin_menu_item("tools",
					 "python-debugger-winpdb",
					 Gio.MenuItem.new(label=_("Python Debugger (winpdb)"),
							  detailed_action="app.python-debugger-winpdb"))

		self.debugpy_action = Gio.SimpleAction.new("python-debugger-debugpy", None)
		self.debugpy_action.connect('activate', self.enable_debugpy, shell)
		app.add_action(self.debugpy_action)

		app.add_plugin_menu_item("tools",
					 "python-debugger-debugpy",
					 Gio.MenuItem.new(label=_("Python Debugger (debugpy)"),
							  detailed_action="app.python-debugger-debugpy"))

	def do_deactivate(self):
		shell = self.object
		app = shell.props.application

		app.remove_plugin_menu_item("tools", "python-console")
		app.remove_action("python-console")

		app.remove_plugin_menu_item("tools", "python-debugger-winpdb")
		app.remove_action("python-debugger-winpdb")
		self.rpdb2_action = None

		app.remove_plugin_menu_item("tools", "python-debugger-debugpy")
		app.remove_action("python-debugger-debugpy")
		self.debugpy_action = None
		
		if self.window is not None:
			self.window.destroy()


	def show_console(self, action, parameter, shell):
		if not self.window:
			ns = {'__builtins__' : __builtins__, 
			      'RB' : RB,
			      'shell' : shell}
			console = PythonConsole(namespace = ns, 
			                        destroy_cb = self.destroy_console)
			console.set_size_request(600, 400)
			console.eval('print("' + \
			             _("You can access the main window " \
			             "through the \'shell\' variable :") +
			             '\\n%s" % shell)', False)

			self.window = Gtk.Window()
			self.window.set_title('Rhythmbox Python Console')
			self.window.add(console)
			self.window.connect('destroy', self.destroy_console)
			self.window.show_all()
		else:
			self.window.show_all()
		self.window.grab_focus()

	def attach_winpdb(self, action, parameter, shell):
		if have_rpdb2 is False:
			self.missing_python_module("rpdb2", "winpdb")
			self.rpdb2_action.set_enabled(False)
			return False

		pwd_path = os.path.join(RB.user_data_dir(), "rpdb2_password")
		msg = _("After you press OK, Rhythmbox will wait until you connect to it with winpdb or rpdb2. If you have not set a debugger password in the file %s, it will use the default password ('rhythmbox').") % pwd_path
		dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.INFO, Gtk.ButtonsType.OK_CANCEL, msg)
		if dialog.run() == Gtk.ResponseType.OK:
			password = "rhythmbox"
			if os.path.exists(pwd_path):
				pwd_file = open(pwd_path)
				password = pwd_file.read().rstrip()
				pwd_file.close()

			def start_debugger(password):
				rpdb2.start_embedded_debugger(password)
				return False

			GLib.idle_add(start_debugger, password)
		dialog.destroy()
		return False
	
	def enable_debugpy(self, action, parameter, shell):
		if have_debugpy is False:
			self.missing_python_module("debugpy", "debugpy")
			self.debugpy_action.set_enabled(False)
			return False
		else:
			try:
				host, port = debugpy.listen(DEBUG_PORT)
				msgtype = Gtk.MessageType.INFO
				msg = _("Rhythmbox is now listening for Debug Adapter Protocol connections on port %d. You can now attach to it using a compatible debugger such as vimspector, nvim-dap or Visual Studio Code.") % port
				self.debugpy_action.set_enabled(False)
			except Exception as e:
				msgtype = Gtk.MessageType.ERROR
				msg = _("Unable to start Debug Adapter Protocol listener: %s") % str(e)

		self.message_dialog(msg, msgtype)
		return False

	def missing_python_module(self, module, debugger):
		msg = _("The %s Python module is not available. Install the module and then restart Rhythmbox to enable debugging with %s.") % (module, debugger)
		self.message_dialog(msg, Gtk.MessageType.ERROR)

	def message_dialog(self, msg, msgtype=Gtk.MessageType.INFO):
		dialog = Gtk.MessageDialog(None, 0, msgtype, Gtk.ButtonsType.OK, msg)
		dialog.connect("response", self.message_dialog_response)
		dialog.show()


	def message_dialog_response(self, dialog, rsp):
		dialog.destroy()

	def destroy_console(self, *args):
		self.window.destroy()
		self.window = None

class PythonConsole(Gtk.ScrolledWindow):

	def get_end_iter(self):
		return self.view.get_buffer().get_end_iter()

	def get_iter_at_mark(self, mark):
		return self.view.get_buffer().get_iter_at_mark(mark)

	def __init__(self, namespace = {}, destroy_cb = None):
		Gtk.ScrolledWindow.__init__(self)

		self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		self.set_shadow_type(Gtk.ShadowType.NONE)
		self.view = Gtk.TextView()
		self.view.modify_font(Pango.font_description_from_string('Monospace'))
		self.view.set_editable(True)
		self.view.set_wrap_mode(Gtk.WrapMode.CHAR)
		self.add(self.view)
		self.view.show()

		buffer = self.view.get_buffer()
		self.normal = buffer.create_tag("normal")
		self.error  = buffer.create_tag("error")
		self.error.set_property("foreground", "red")
		self.command = buffer.create_tag("command")
		self.command.set_property("foreground", "blue")

		self.__spaces_pattern = re.compile(r'^\s+')		
		self.namespace = namespace

		self.destroy_cb = destroy_cb

		self.block_command = False

		# Init first line
		buffer.create_mark("input-line", self.get_end_iter(), True)
		buffer.insert(self.get_end_iter(), ">>> ")
		buffer.create_mark("input", self.get_end_iter(), True)

		# Init history
		self.history = ['']
		self.history_pos = 0
		self.current_command = ''
		self.namespace['__history__'] = self.history

		# Set up hooks for standard output.
		self.stdout = gtkoutfile(self, sys.stdout.fileno(), self.normal)
		self.stderr = gtkoutfile(self, sys.stderr.fileno(), self.error)

		# Signals
		self.view.connect("key-press-event", self.__key_press_event_cb)
		buffer.connect("mark-set", self.__mark_set_cb)


	def __key_press_event_cb(self, view, event):
		if event.keyval == Gdk.KEY_D and \
		   event.state == Gdk.ModifierType.CONTROL_MASK:
			self.destroy()
		
		elif event.keyval == Gdk.KEY_Return and \
		     event.state == Gdk.ModifierType.CONTROL_MASK:
			# Get the command
			buffer = view.get_buffer()
			inp_mark = buffer.get_mark("input")
			inp = self.get_iter_at_mark(inp_mark)
			cur = self.get_end_iter()
			line = buffer.get_text(inp, cur, True)
			self.current_command = self.current_command + line + "\n"
			self.history_add(line)

			# Prepare the new line
			cur = self.get_end_iter()
			buffer.insert(cur, "\n... ")
			cur = self.get_end_iter()
			buffer.move_mark(inp_mark, cur)
			
			# Keep indentation of precendent line
			spaces = re.match(self.__spaces_pattern, line)
			if spaces is not None:
				buffer.insert(cur, line[spaces.start() : spaces.end()])
				cur = self.get_end_iter()
				
			buffer.place_cursor(cur)
			GLib.idle_add(self.scroll_to_end)
			return True
		
		elif event.keyval == Gdk.KEY_Return:
			# Get the marks
			buffer = view.get_buffer()
			lin_mark = buffer.get_mark("input-line")
			inp_mark = buffer.get_mark("input")

			# Get the command line
			inp = self.get_iter_at_mark(inp_mark)
			cur = self.get_end_iter()
			line = buffer.get_text(inp, cur, True)
			self.current_command = self.current_command + line + "\n"
			self.history_add(line)

			# Make the line blue
			lin = self.get_iter_at_mark(lin_mark)
			buffer.apply_tag(self.command, lin, cur)
			buffer.insert(cur, "\n")
			
			cur_strip = self.current_command.rstrip()

			if cur_strip.endswith(":") \
			or (self.current_command[-2:] != "\n\n" and self.block_command):
				# Unfinished block command
				self.block_command = True
				com_mark = "... "
			elif cur_strip.endswith("\\"):
				com_mark = "... "
			else:
				# Eval the command
				self.__run(self.current_command)
				self.current_command = ''
				self.block_command = False
				com_mark = ">>> "

			# Prepare the new line
			cur = self.get_end_iter()
			buffer.move_mark(lin_mark, cur)
			buffer.insert(cur, com_mark)
			cur = self.get_end_iter()
			buffer.move_mark(inp_mark, cur)
			buffer.place_cursor(cur)
			GLib.idle_add(self.scroll_to_end)
			return True

		elif event.keyval == Gdk.KEY_KP_Down or \
		     event.keyval == Gdk.KEY_Down:
			# Next entry from history
			view.emit_stop_by_name("key_press_event")
			self.history_down()
			GLib.idle_add(self.scroll_to_end)
			return True

		elif event.keyval == Gdk.KEY_KP_Up or \
		     event.keyval == Gdk.KEY_Up:
			# Previous entry from history
			view.emit_stop_by_name("key_press_event")
			self.history_up()
			GLib.idle_add(self.scroll_to_end)
			return True

		elif event.keyval == Gdk.KEY_KP_Left or \
		     event.keyval == Gdk.KEY_Left or \
		     event.keyval == Gdk.KEY_BackSpace:
			buffer = view.get_buffer()
			inp = self.get_iter_at_mark(buffer.get_mark("input"))
			cur = self.get_iter_at_mark(buffer.get_insert())
			return inp.compare(cur) == 0

		elif event.keyval == Gdk.KEY_Home:
			# Go to the begin of the command instead of the begin of
			# the line
			buffer = view.get_buffer()
			inp = self.get_iter_at_mark(buffer.get_mark("input"))
			if event.state == Gdk.ModifierType.SHIFT_MASK:
				buffer.move_mark_by_name("insert", inp)
			else:
				buffer.place_cursor(inp)
			return True
		
	def __mark_set_cb(self, buffer, iter, name):
		input = self.get_iter_at_mark(buffer.get_mark("input"))
		pos   = self.get_iter_at_mark(buffer.get_insert())
		self.view.set_editable(pos.compare(input) != -1)

	def get_command_line(self):
		buffer = self.view.get_buffer()
		inp = self.get_iter_at_mark(buffer.get_mark("input"))
		cur = self.get_end_iter()
		return buffer.get_text(inp, cur, True)
	
	def set_command_line(self, command):
		buffer = self.view.get_buffer()
		mark = buffer.get_mark("input")
		inp = self.get_iter_at_mark(mark)
		cur = self.get_end_iter()
		buffer.delete(inp, cur)
		buffer.insert(inp, command)
		buffer.select_range(self.get_iter_at_mark(mark),
		                    self.get_end_iter())
		self.view.grab_focus()
	
	def history_add(self, line):
		if line.strip() != '':
			self.history_pos = len(self.history)
			self.history[self.history_pos - 1] = line
			self.history.append('')
	
	def history_up(self):
		if self.history_pos > 0:
			self.history[self.history_pos] = self.get_command_line()
			self.history_pos = self.history_pos - 1
			self.set_command_line(self.history[self.history_pos])
			
	def history_down(self):
		if self.history_pos < len(self.history) - 1:
			self.history[self.history_pos] = self.get_command_line()
			self.history_pos = self.history_pos + 1
			self.set_command_line(self.history[self.history_pos])
	
	def scroll_to_end(self):
		iter = self.get_end_iter()
		self.view.scroll_to_iter(iter, 0.0, False, 0.5, 0.5)
		return False

	def write(self, text, tag = None):
		buffer = self.view.get_buffer()
		if tag is None:
			buffer.insert(self.get_end_iter(), text)
		else:
			iter = self.get_end_iter()
			offset = iter.get_offset()
			buffer.insert(iter, text)

			start_iter = Gtk.TextIter()
			start_iter = buffer.get_iter_at_offset(offset)
			buffer.apply_tag(tag, start_iter, self.get_end_iter())
		GLib.idle_add(self.scroll_to_end)

	def eval(self, command, display_command = False):
		buffer = self.view.get_buffer()
		lin = buffer.get_mark("input-line")
		buffer.delete(self.get_iter_at_mark(lin),
			      self.get_end_iter())

		if isinstance(command, list) or isinstance(command, tuple):
			for c in command:
				if display_command:
					self.write(">>> " + c + "\n", self.command)
				self.__run(c)
		else:
			if display_command:
				self.write(">>> " + c + "\n", self.command)
			self.__run(command)

		cur = self.get_end_iter()
		buffer.move_mark_by_name("input-line", cur)
		buffer.insert(cur, ">>> ")
		cur = self.get_end_iter()
		buffer.move_mark_by_name("input", cur)
		self.view.scroll_to_iter(self.get_end_iter(), 0.0, False, 0.5, 0.5)
	
	def __run(self, command):
		sys.stdout, self.stdout = self.stdout, sys.stdout
		sys.stderr, self.stderr = self.stderr, sys.stderr
		
		try:
			try:
				r = eval(command, self.namespace, self.namespace)
				if r is not None:
					print(repr(r))
			except SyntaxError:
				exec(command, self.namespace)
		except:
			if hasattr(sys, 'last_type') and sys.last_type == SystemExit:
				self.destroy()
			else:
				traceback.print_exc()
		finally:
			sys.stdout, self.stdout = self.stdout, sys.stdout
			sys.stderr, self.stderr = self.stderr, sys.stderr


	def destroy(self):
		if self.destroy_cb is not None:
			self.destroy_cb()
		
class gtkoutfile:
	"""A fake output file object.  It sends output to a TK test widget,
	and if asked for a file number, returns one set on instance creation"""
	def __init__(self, console, fn, tag):
		self.fn = fn
		self.console = console
		self.tag = tag
	def close(self):         pass
	def flush(self):         pass
	def fileno(self):        return self.fn
	def isatty(self):        return 0
	def read(self, a):       return ''
	def readline(self):      return ''
	def readlines(self):     return []
	def write(self, s):      self.console.write(s, self.tag)
	def writelines(self, l): self.console.write(l, self.tag)
	def seek(self, a):       raise IOError((29, 'Illegal seek'))
	def tell(self):          raise IOError((29, 'Illegal seek'))
	truncate = tell

# ex:noet:ts=8:



r'''
<MIT License>
Copyright (c) 2013  Marek Majkowski <marek@popcount.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
</MIT License>


SipHash-2-4 implementation, following the 'hashlib' API:

>>> key = b'0123456789ABCDEF'
>>> SipHash_2_4(key, b'a').hexdigest()
b'864c339cb0dc0fac'
>>> SipHash_2_4(key, b'a').digest()
b'\x86L3\x9c\xb0\xdc\x0f\xac'
>>> SipHash_2_4(key, b'a').hash()
12398370950267227270
>>> SipHash_2_4(key).update(b'a').hash()
12398370950267227270

>>> key = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'
>>> SipHash_2_4(key, b'').hash()
8246050544436514353
>>> SipHash_2_4(key, b'').hexdigest()
b'310e0edd47db6f72'

'''
import struct
import binascii

def _doublesipround(v, m):
    '''
    Internal helper. Xors 'm' to 'v3', runs two rounds of siphash on
    vector 'v' and xors 'm' to 'v0'.

    >>> _doublesipround((1,2,3,4),0)
    (9263201270060220426, 2307743542053503000, 5255419393243893904, 10208987565802066018)
    >>> _doublesipround((1,2,3,4),0xff)
    (11557575153743626750, 2307780510495171325, 7519994316568162407, 5442382437785464174)
    >>> _doublesipround((0,0,0,0),0)
    (0, 0, 0, 0)
    >>> _doublesipround((0,0,0,0),0xff)
    (2368684213854535680, 36416423977725, 2305811110491594975, 15626573430810475768)
    '''
    a, b, c, d = v
    d ^= m

    e = (a + b) & 0xffffffffffffffff
    i = (((b & 0x7ffffffffffff) << 13) | (b >> 51)) ^ e
    f = c + d
    j = ((((d) << 16) | (d >> 48)) ^ f ) & 0xffffffffffffffff
    h = (f + i) & 0xffffffffffffffff

    k = ((e << 32) | (e >> 32)) + j
    l = (((i & 0x7fffffffffff) << 17) | (i >> 47)) ^ h
    o = (((j << 21) | (j >> 43)) ^ k) & 0xffffffffffffffff

    p = (k + l) & 0xffffffffffffffff
    q = (((l & 0x7ffffffffffff) << 13) | (l >> 51)) ^ p
    r = ((h << 32) | (h >> 32)) + o
    s = (((o << 16) | (o >> 48)) ^ r) & 0xffffffffffffffff
    t = (r + q) & 0xffffffffffffffff
    u = (((p << 32) | (p >> 32)) + s) & 0xffffffffffffffff

    return (u ^ m,
            (((q & 0x7fffffffffff) << 17) | (q >> 47)) ^ t,
            ((t & 0xffffffff) << 32) | (t >> 32),
            (((s & 0x7ffffffffff) << 21) | (s >> 43)) ^ u)


_zeroes = b'\x00\x00\x00\x00\x00\x00\x00\x00'
_oneQ = struct.Struct('<Q')
_twoQ = struct.Struct('<QQ')
_oneQout = struct.Struct(">Q")


class SipHash_2_4(object):
    r'''
    >>> SipHash_2_4(b'0123456789ABCDEF', b'a').hash()
    12398370950267227270
    >>> SipHash_2_4(b'0123456789ABCDEF', b'').hash()
    3627314469837380007
    >>> SipHash_2_4(b'FEDCBA9876543210', b'').hash()
    2007056766899708634
    >>> SipHash_2_4(b'FEDCBA9876543210').update(b'').update(b'').hash()
    2007056766899708634
    >>> SipHash_2_4(b'FEDCBA9876543210', b'a').hash()
    6581475155582014123
    >>> SipHash_2_4(b'FEDCBA9876543210').update(b'a').hash()
    6581475155582014123
    >>> SipHash_2_4(b'FEDCBA9876543210').update(b'a').update(b'').hash()
    6581475155582014123
    >>> SipHash_2_4(b'FEDCBA9876543210').update(b'').update(b'a').hash()
    6581475155582014123

    >>> a = SipHash_2_4(b'FEDCBA9876543210').update(b'a')
    >>> a.hash()
    6581475155582014123
    >>> b = a.copy()
    >>> a.hash(), b.hash()
    (6581475155582014123, 6581475155582014123)
    >>> a.update(b'a') and None
    >>> a.hash(), b.hash()
    (3258273892680892829, 6581475155582014123)
    '''
    digest_size = 16
    block_size = 64

    s = b''
    b = 0

    def __init__(self, secret, s=b''):
        k0 = (secret[1] << 32) + secret[0]
        k1 = (secret[3] << 32) + secret[2]
        self.v = (0x736f6d6570736575 ^ k0,
                  0x646f72616e646f6d ^ k1,
                  0x6c7967656e657261 ^ k0,
                  0x7465646279746573 ^ k1)
        self.update(s)

    def update(self, s):
        s = self.s + s
        lim = (len(s)//8)*8
        v = self.v
        off = 0

        for off in range(0, lim, 8):
            m, = _oneQ.unpack_from(s, off)

            # print 'v0 %016x' % v[0]
            # print 'v1 %016x' % v[1]
            # print 'v2 %016x' % v[2]
            # print 'v3 %016x' % v[3]
            # print 'compress %016x' % m

            v = _doublesipround(v, m)
        self.v = v
        self.b += lim
        self.s = s[lim:]
        return self

    def hash(self):
        l = len(self.s)
        assert l < 8

        b = (((self.b + l) & 0xff) << 56)
        b |= _oneQ.unpack_from(self.s+_zeroes)[0]
        v = self.v

        # print 'v0 %016x' % v[0]
        # print 'v1 %016x' % v[1]
        # print 'v2 %016x' % v[2]
        # print 'v3 %016x' % v[3]
        # print 'padding %016x' % b

        v = _doublesipround(v, b)

        # print 'v0 %016x' % v0
        # print 'v1 %016x' % v1
        # print 'v2 %016x' % v2
        # print 'v3 %016x' % v3

        v = list(v)
        v[2] ^= 0xff
        v = _doublesipround(_doublesipround(v, 0), 0)
        return v[0] ^ v[1] ^ v[2] ^ v[3]

    def digest(self):
        return _oneQout.pack(self.hash())

    def hexdigest(self):
        return binascii.hexlify(self.digest())

    def copy(self):
        n = SipHash_2_4(_zeroes * 2)
        n.v, n.s, n.b = self.v, self.s, self.b
        return n


siphash24 = SipHash_2_4
SipHash24 = SipHash_2_4


if __name__ == "__main__":
    # Test vectors as per spec
    vectors = [c.encode('utf-8') for c in [
        "310e0edd47db6f72", "fd67dc93c539f874", "5a4fa9d909806c0d", "2d7efbd796666785",
        "b7877127e09427cf", "8da699cd64557618", "cee3fe586e46c9cb", "37d1018bf50002ab",
        "6224939a79f5f593", "b0e4a90bdf82009e", "f3b9dd94c5bb5d7a", "a7ad6b22462fb3f4",
        "fbe50e86bc8f1e75", "903d84c02756ea14", "eef27a8e90ca23f7", "e545be4961ca29a1",
        "db9bc2577fcc2a3f", "9447be2cf5e99a69", "9cd38d96f0b3c14b", "bd6179a71dc96dbb",
        "98eea21af25cd6be", "c7673b2eb0cbf2d0", "883ea3e395675393", "c8ce5ccd8c030ca8",
        "94af49f6c650adb8", "eab8858ade92e1bc", "f315bb5bb835d817", "adcf6b0763612e2f",
        "a5c91da7acaa4dde", "716595876650a2a6", "28ef495c53a387ad", "42c341d8fa92d832",
        "ce7cf2722f512771", "e37859f94623f3a7", "381205bb1ab0e012", "ae97a10fd434e015",
        "b4a31508beff4d31", "81396229f0907902", "4d0cf49ee5d4dcca", "5c73336a76d8bf9a",
        "d0a704536ba93e0e", "925958fcd6420cad", "a915c29bc8067318", "952b79f3bc0aa6d4",
        "f21df2e41d4535f9", "87577519048f53a9", "10a56cf5dfcd9adb", "eb75095ccd986cd0",
        "51a9cb9ecba312e6", "96afadfc2ce666c7", "72fe52975a4364ee", "5a1645b276d592a1",
        "b274cb8ebf87870a", "6f9bb4203de7b381", "eaecb2a30b22a87f", "9924a43cc1315724",
        "bd838d3aafbf8db7", "0b1a2a3265d51aea", "135079a3231ce660", "932b2846e4d70666",
        "e1915f5cb1eca46c", "f325965ca16d629f", "575ff28e60381be5", "724506eb4c328a95",
        ]]

    key = ''.join(chr(i) for i in range(16)).encode('utf-8')
    plaintext = ''.join(chr(i) for i in range(64)).encode('utf-8')
    for i in range(64):
        assert SipHash_2_4(key, plaintext[:i]).hexdigest() == vectors[i], \
            'failed on test no %i' % i

    # Internal doctests
    #
    # To maintain compatibility with both python 2.x and 3.x in tests
    # we need to do a trick. Python 2.x doesn't like b'' notation,
    # Python 3.x doesn't have 2222L long integers notation. To
    # overcome that we'll pipe both results as well as the intended
    # doctest output through an `eval` function before comparison. To
    # do it we need to monkeypatch the OutputChecker:
    import doctest
    EVAL_FLAG = doctest.register_optionflag("EVAL")
    OrigOutputChecker = doctest.OutputChecker

    def relaxed_eval(s):
        if s.strip():
            return eval(s)
        else:
            return None

    class MyOutputChecker:
        def __init__(self):
            self.orig = OrigOutputChecker()

        def check_output(self, want, got, optionflags):
            if optionflags & EVAL_FLAG:
                return relaxed_eval(got) == relaxed_eval(want)
            else:
                return self.orig.check_output(want, got, optionflags)

        def output_difference(self, example, got, optionflags):
            return self.orig.output_difference(example, got, optionflags)

    doctest.OutputChecker = MyOutputChecker
    # Monkey patching done. Go for doctests:

    if doctest.testmod(optionflags=EVAL_FLAG)[0] == 0: print("all tests ok")



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2006 Adam Zimmerman  <adam_zimmerman@sfu.ca>
# Copyright (C) 2006 James Livingston  <doclivingston@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import os
import sys
import xml
import urllib.parse, urllib.request
import threading
import zipfile

import rb
from gi.repository import RB
from gi.repository import GObject, Gtk, Gdk, Gio, GLib

from TrackListHandler import TrackListHandler
from DownloadAlbumHandler import DownloadAlbumHandler, MagnatuneDownloadError
import MagnatuneAccount

import gettext
gettext.install('rhythmbox', RB.locale_dir())

magnatune_partner_id = "rhythmbox"

# URIs
magnatune_song_info_uri = "http://magnatune.com/info/song_info_xml.zip"
magnatune_changed_uri = "http://magnatune.com/info/changed.txt"
magnatune_member_signup_uri = "http://magnatune.com/member/signup?"
magnatune_api_download_uri = "http://%s:%s@download.magnatune.com/buy/membership_free_dl_xml?"

magnatune_in_progress_dir = Gio.file_new_for_path(RB.user_data_dir()).resolve_relative_path('magnatune')
magnatune_cache_dir = Gio.file_new_for_path(RB.user_cache_dir()).resolve_relative_path('magnatune')

magnatune_song_info = os.path.join(magnatune_cache_dir.get_path(), 'song_info.xml')
magnatune_song_info_temp = os.path.join(magnatune_cache_dir.get_path(), 'song_info.zip.tmp')
magnatune_changes = os.path.join(magnatune_cache_dir.get_path(), 'changed.txt')


class MagnatuneSource(RB.BrowserSource):
	def __init__(self):
		RB.BrowserSource.__init__(self)
		self.hate = self

		self.__popup = None
		self.__settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.magnatune")
		# source state
		self.__activated = False
		self.__db = None
		self.__info_screen = None # the loading screen

		# track data
		self.__sku_dict = {}
		self.__home_dict = {}
		self.__art_dict = {}

		# catalogue stuff
		self.__has_loaded = False # whether the catalog has been loaded yet
		self.__update_id = 0 # GLib.idle_add id for catalog updates
		self.__catalogue_loader = None
		self.__catalogue_check = None
		self.__load_progress = None
		self.__download_progress = None

		# album download stuff
		self.__downloads = {} # keeps track of download progress for each file
		self.__copies = {} # keeps copy objects for each file

		self.__art_store = RB.ExtDB(name="album-art")

	#
	# RBSource methods
	#

	def do_show_entry_popup(self):
		if self.__popup is None:
			builder = Gtk.Builder()
			builder.add_from_file(rb.find_plugin_file(self.props.plugin, "magnatune-popup.ui"))
			self.__popup = builder.get_object("magnatune-popup")

		menu = Gtk.Menu.new_from_model(self.__popup)
		menu.attach_to_widget(self, None)
		menu.popup(None, None, None, None, 3, Gtk.get_current_event_time())


	def do_selected(self):
		if not self.__activated:
			shell = self.props.shell
			self.__db = shell.props.db
			self.__entry_type = self.props.entry_type

			if not magnatune_in_progress_dir.query_exists(None):
				magnatune_in_progress_path = magnatune_in_progress_dir.get_path()
				os.mkdir(magnatune_in_progress_path, 0o700)

			if not magnatune_cache_dir.query_exists(None):
				magnatune_cache_path = magnatune_cache_dir.get_path()
				os.mkdir(magnatune_cache_path, 0o700)

			self.__activated = True
			self.__show_loading_screen(True)

			# start our catalogue updates
			self.__update_id = GLib.timeout_add_seconds(6 * 60 * 60, self.__update_catalogue)
			self.__update_catalogue()

	def do_can_delete(self):
		return False

	def do_pack_content(self, content):
		self.__paned_box = Gtk.VBox(homogeneous=False, spacing=5)
		self.pack_start(self.__paned_box, True, True, 0)
		self.__paned_box.pack_start(content, True, True, 0)


	def do_delete_thyself(self):
		if self.__update_id != 0:
			GLib.source_remove(self.__update_id)
			self.__update_id = 0

		if self.__catalogue_loader is not None:
			self.__catalogue_loader.cancel()
			self.__catalogue_loader = None

		if self.__catalogue_check is not None:
			self.__catalogue_check.cancel()
			self.__catalogue_check = None

		RB.BrowserSource.do_delete_thyself(self)

	#
	# methods for use by plugin and UI
	#

	def display_artist_info(self):
		screen = self.props.shell.props.window.get_screen()
		tracks = self.get_entry_view().get_selected_entries()
		if len(tracks) == 0:
			return

		tr = tracks[0]
		sku = self.__sku_dict[tr.get_string(RB.RhythmDBPropType.LOCATION)]
		url = self.__home_dict[sku]
		Gtk.show_uri(screen, url, Gdk.CURRENT_TIME)


	def download_redirect(self):
		screen = self.props.shell.props.window.get_screen()
		tracks = self.get_entry_view().get_selected_entries()
		if len(tracks) == 0:
			return

		url = magnatune_member_signup_uri + urllib.parse.urlencode({ 'ref': magnatune_partner_id })
		Gtk.show_uri(screen, url, Gdk.CURRENT_TIME)


	def download_album(self):
		if self.__settings['account-type'] != 'download':
			# The user doesn't have a download account, so redirect them to the member signup page
			self.download_redirect()
			return

		try:
			# Just use the first library location
			library = Gio.Settings.new("org.gnome.rhythmbox.rhythmdb")
			library_location = library['locations'][0]
		except IndexError as e:
			RB.error_dialog(title = _("Couldn't download album"),
				        message = _("You must have a library location set to download an album."))
			return

		tracks = self.get_entry_view().get_selected_entries()
		skus = []

		for track in tracks:
			sku = self.__sku_dict[track.get_string(RB.RhythmDBPropType.LOCATION)]
			if sku in skus:
				continue
			skus.append(sku)
			self.__auth_download(sku)

	#
	# internal catalogue downloading and loading
	#

	def __update_catalogue(self):
		def update_cb(remote_changes):
			self.__catalogue_check = None
			try:
				f = open(magnatune_changes, 'rt')
				local_changes = f.read().strip()
			except:
				local_changes = ""

			remote_changes = remote_changes.strip().decode("iso-8859-1")
			print("local checksum %s, remote checksum %s" % (local_changes, remote_changes))
			if local_changes != remote_changes:
				try:
					f = open(magnatune_changes, 'wt')
					f.write(remote_changes + "\n")
					f.close()
				except Exception as e:
					print("unable to write local change id: %s" % str(e))

				download_catalogue()
			elif self.__has_loaded is False:
				load_catalogue()

		def download_catalogue():
			def find_song_info(catalogue):
				for info in catalogue.infolist():
					if info.filename.endswith("song_info.xml"):
						return info.filename;
				return None

			def download_progress(copy, complete, total, self):
				self.__load_progress.props.task_progress = min(float(complete) / total, 1.0)

			def download_finished(copy, success, self):
				if not success:
					print("catalog download failed")
					print(copy.get_error())
					return

				print("catalog download successful")
				# done downloading, unzip to real location
				catalog_zip = zipfile.ZipFile(magnatune_song_info_temp)
				catalog = open(magnatune_song_info, 'wb')
				filename = find_song_info(catalog_zip)
				if filename is None:
					RB.error_dialog(title=_("Unable to load catalog"),
							message=_("Rhythmbox could not understand the Magnatune catalog, please file a bug."))
					return
				catalog.write(catalog_zip.read(filename))
				catalog.close()
				catalog_zip.close()

				df = Gio.file_new_for_path(magnatune_song_info_temp)
				df.delete(None)
				self.__catalogue_loader = None

				self.__load_progress.props.task_outcome = RB.TaskOutcome.COMPLETE

				load_catalogue()

			try:
				df = Gio.file_new_for_path(magnatune_song_info_temp)
				df.delete(None)
			except:
				pass

			self.__load_progress = RB.TaskProgressSimple.new()
			self.__load_progress.props.task_label = _("Loading Magnatune catalog")
			self.props.shell.props.task_list.add_task(self.__load_progress)

			self.__catalog_loader = RB.AsyncCopy()
			self.__catalog_loader.set_progress(download_progress, self)
			self.__catalog_loader.start(magnatune_song_info_uri, magnatune_song_info_temp, download_finished, self)

		def load_catalogue():

			def catalogue_chunk_cb(loader, chunk, total, parser):
				if chunk is None:
					self.__load_progress.props.task_outcome = RB.TaskOutcome.COMPLETE
					error = loader.get_error()
					if error:
						# report error somehow?
						print("error loading catalogue: %s" % error)

					try:
						parser.close()
					except xml.sax.SAXParseException as e:
						# there isn't much we can do here
						print("error parsing catalogue: %s" % e)

					self.__show_loading_screen(False)
					self.__catalogue_loader = None

					# restart in-progress downloads
					# (doesn't really belong here)
					for f in magnatune_in_progress_dir.enumerate_children('standard::name', Gio.FileQueryInfoFlags.NONE, None):
						name = f.get_name()
						if not name.startswith("in_progress_"):
							continue
						(result, uri, etag) = magnatune_in_progress_dir.resolve_relative_path(name).load_contents(None)
						uri = uri.decode('utf-8')

						print("restarting download from %s" % uri)
						self.__download_album(uri, name[12:])
				else:
					# hack around some weird chars that show up in the catalogue for some reason
					data = chunk.get_data().decode('utf-8', errors='replace')
					data = data.replace("\x19", "'")
					data = data.replace("\x13", "-")

					# argh.
					data = data.replace("Rock & Roll", "Rock &amp; Roll")

					try:
						parser.feed(data)
					except xml.sax.SAXParseException as e:
						print("error parsing catalogue: %s" % e)

					load_size['size'] += len(data)
					self.__load_progress.props.task_progress = min(float(load_size['size']) / total, 1.0)


			self.__has_loaded = True
			self.__load_progress = RB.TaskProgressSimple.new()
			self.__load_progress.props.task_label = _("Loading Magnatune catalog")
			self.props.shell.props.task_list.add_task(self.__load_progress)

			load_size = {'size': 0}

			parser = xml.sax.make_parser()
			parser.setContentHandler(TrackListHandler(self.__db, self.__entry_type, self.__sku_dict, self.__home_dict, self.__art_dict))

			self.__catalogue_loader = RB.ChunkLoader()
			self.__catalogue_loader.set_callback(catalogue_chunk_cb, parser)
			self.__catalogue_loader.start(magnatune_song_info, 64*1024)


		self.__catalogue_check = rb.Loader()
		self.__catalogue_check.get_url(magnatune_changed_uri, update_cb)


	def __show_loading_screen(self, show):
		if self.__info_screen is None:
			# load the builder stuff
			builder = Gtk.Builder()
			builder.add_from_file(rb.find_plugin_file(self.props.plugin, "magnatune-loading.ui"))
			self.__info_screen = builder.get_object("magnatune_loading_scrolledwindow")
			self.pack_start(self.__info_screen, True, True, 0)
			self.get_entry_view().set_no_show_all(True)
			self.__info_screen.set_no_show_all(True)

		self.__info_screen.set_property("visible", show)
		self.__paned_box.set_property("visible", not show)

	def __notify_status_changed(self):
		pass

	#
	# internal purchasing code
	#

	def __auth_download(self, sku): # http://magnatune.com/info/api

		def auth_data_cb(data, userpass):
			(username, password) = userpass
			dl_album_handler = DownloadAlbumHandler(self.__settings['format'])
			auth_parser = xml.sax.make_parser()
			auth_parser.setContentHandler(dl_album_handler)

			if data is None:
				# hmm.
				return

			try:
				data = data.decode("utf-8")
				data = data.replace("<br>", "") # get rid of any stray <br> tags that will mess up the parser
				data = data.replace(" & ", " &amp; ") # clean up some missing escaping
				# print data
				auth_parser.feed(data)
				auth_parser.close()

				# process the URI: add authentication info, quote the filename component for some reason
				parsed = urllib.parse.urlparse(dl_album_handler.url)
				netloc = "%s:%s@%s" % (username, password, parsed.hostname)

				spath = os.path.split(urllib.request.url2pathname(parsed.path))
				basename = spath[1]
				path = urllib.request.pathname2url(os.path.join(spath[0], urllib.parse.quote(basename)))

				authed = (parsed[0], netloc, path) + parsed[3:]
				audio_dl_uri = urllib.parse.urlunparse(authed)

				print("download uri for %s is %s" % (sku, audio_dl_uri))
				self.__download_album(audio_dl_uri, sku)

			except MagnatuneDownloadError as e:
				RB.error_dialog(title = _("Download Error"),
						message = _("An error occurred while trying to authorize the download.\nThe Magnatune server returned:\n%s") % str(e))
			except Exception as e:
				sys.excepthook(*sys.exc_info())
				RB.error_dialog(title = _("Error"),
						message = _("An error occurred while trying to download the album.\nThe error text is:\n%s") % str(e))

		print("downloading album: " + sku)
		account = MagnatuneAccount.instance()
		(account_type, username, password) = account.get()
		url_dict = {
			'id':	magnatune_partner_id,
			'sku':	sku
		}
		url = magnatune_api_download_uri % (username, password)
		url = url + urllib.parse.urlencode(url_dict)

		l = rb.Loader()
		l.get_url(url, auth_data_cb, (username, password))


	def __download_album(self, audio_dl_uri, sku):
		def update_progress(self):
			if len(self.__downloads) == 0:
				self.__download_progress.props.task_outcome = RB.TaskOutcome.COMPLETE
				self.__download_progress = None
			else:
				complete, total = map(sum, zip(*self.__downloads.values()))
				if total > 0:
					self.__download_progress.props.task_progress = min(float(complete) / total, 1.0)

		def download_progress(copy, complete, total, self):
			self.__downloads[audio_dl_uri] = (complete, total)
			update_progress(self)

		def download_finished(copy, success, self):
			del self.__downloads[audio_dl_uri]
			del self.__copies[audio_dl_uri]

			print("download of %s finished: %s" % (audio_dl_uri, success))
			if success:
				threading.Thread(target=unzip_album).start()
			else:
				remove_download_files()

			update_progress(self)


		def unzip_album():
			# just use the first library location
			library = Gio.Settings.new("org.gnome.rhythmbox.rhythmdb")
			library_location = Gio.file_new_for_uri(library['locations'][0])

			print("unzipping %s" % dest.get_path())
			album = zipfile.ZipFile(dest.get_path())
			for track in album.namelist():
				track_uri = library_location.resolve_relative_path(track).get_uri()
				print("zip file entry: %s => %s" % (track, track_uri))

				track_uri = RB.sanitize_uri_for_filesystem(track_uri)
				RB.uri_create_parent_dirs(track_uri)

				track_out = Gio.file_new_for_uri(track_uri).replace(None, False, Gio.FileCreateFlags.NONE, None)
				if track_out is not None:
					track_out.write(album.read(track), None)
					track_out.close(None)
					print("adding %s to library" % track_uri)
					self.__db.add_uri(track_uri)

			album.close()
			remove_download_files()

		def remove_download_files():
			print("removing download files")
			in_progress.delete(None)
			dest.delete(None)

		in_progress = magnatune_in_progress_dir.resolve_relative_path("in_progress_" + sku)
		dest = magnatune_in_progress_dir.resolve_relative_path(sku)

		in_progress.replace_contents(audio_dl_uri.encode('utf-8'),
					     None,
					     False,
					     Gio.FileCreateFlags.PRIVATE|Gio.FileCreateFlags.REPLACE_DESTINATION,
					     None)

		if self.__download_progress is None:
			self.__download_progress = RB.TaskProgressSimple.new()
			self.__download_progress.props.task_label = _("Downloading from Magnatune")
			self.__download_progress.connect('cancel-task', self.cancel_downloads)
			self.props.shell.props.task_list.add_task(self.__download_progress)

		# one async copy object per download
		if audio_dl_uri not in self.__copies:
			dl = RB.AsyncCopy()
			dl.set_progress(download_progress, self)
			dl.start(audio_dl_uri, dest.get_uri(), download_finished, self)
			self.__downloads[audio_dl_uri] = (0, 0) # (current, total)
			self.__copies[audio_dl_uri] = dl

	def cancel_downloads(self, task):
		for download in self.__copies.values():
			download.cancel()

		task.props.task_outcome = RB.TaskOutcome.CANCELLED

	def playing_entry_changed(self, entry):
		if not self.__db or not entry:
			return
		if entry.get_entry_type() != self.__db.entry_type_get_by_name("magnatune"):
			return

		sku = self.__sku_dict[entry.get_string(RB.RhythmDBPropType.LOCATION)]
		key = RB.ExtDBKey.create_storage("album", entry.get_string(RB.RhythmDBPropType.ALBUM))
		key.add_field("artist", entry.get_string(RB.RhythmDBPropType.ARTIST))
		self.__art_store.store_uri(key, RB.ExtDBSourceType.EMBEDDED, self.__art_dict[sku])

GObject.type_register(MagnatuneSource)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright 2006, James Livingston <doclivingston@gmail.com>
# Copyright 2006, Ed Catmur <ed@catmur.co.uk>
# Copyright 2007, Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import sys
import os.path
import os
import time

from gi.repository import RB, Gtk

# rb classes
from Loader import Loader
from Coroutine import Coroutine
from URLCache import URLCache
import rbconfig

def try_load_icon(theme, icon, size, flags):
	try:
		return theme.load_icon(icon, size, flags)
	except:
		return None

def append_plugin_source_path(plugin, iconpath):
	theme = Gtk.IconTheme.get_default()

	# get plugin data dir
	datadir = plugin.plugin_info.get_data_dir()
	icondir = os.path.join(datadir, iconpath)
	if os.path.exists(icondir):
		theme.append_search_path(icondir)

	# where was the caller loaded from?
	fr = sys._getframe(1)
	co = fr.f_code
	filename = co.co_filename

	# if the calling plugin has an icons dir, add it to the search path
	plugindir = filename[:filename.rfind(os.sep)]
	icondir = os.path.join(plugindir, iconpath)
	if os.path.exists(icondir):
		theme.append_search_path(icondir)

def entry_equal(a, b):
	if (a is None and b is None):
		return True
	if (a is None or b is None):
		return False
	return a.get_string(RB.RhythmDBPropType.LOCATION) == b.get_string(RB.RhythmDBPropType.LOCATION)

def find_plugin_file(plugin, filename):
	info = plugin.plugin_info
	data_dir = info.get_data_dir()
	path = os.path.join(data_dir, filename)
	print("looking for " + filename + " in " + data_dir)
	if os.path.exists(path):
		return path

	return RB.file(filename)


class _rbdebugfile:
	def __init__(self, fn):
		self.fn = fn

	def write(self, data):
		if data == '\n':
			return
		fr = sys._getframe(1)

		co = fr.f_code
		filename = co.co_filename

		# strip off the cwd, for if running uninstalled
		cwd = os.getcwd()
		if cwd[-1] != os.sep:
			cwd += os.sep
		if filename[:len(cwd)] == cwd:
			filename = filename[len(cwd):]

		# add the class name to the method, if 'self' exists
		methodname = co.co_name
		if 'self' in fr.f_locals:
			methodname = '%s.%s' % (fr.f_locals['self'].__class__.__name__, methodname)

		ln = co.co_firstlineno + fr.f_lineno
		RB.debug_real (methodname, filename, ln, True, str(data))

	def close(self):         pass
	def flush(self):         pass
	def fileno(self):        return self.fn
	def isatty(self):        return 0
	def read(self, a):       return ''
	def readline(self):      return ''
	def readlines(self):     return []
	writelines = write
	def seek(self, a):       raise IOError((29, 'Illegal seek'))
	def tell(self):          raise IOError((29, 'Illegal seek'))
	truncate = tell

sys.stdout = _rbdebugfile(sys.stdout.fileno())



# -.- coding: utf-8 -.-

# Copyright © 2009 Markus Korn <thekorn@gmx.de>
# Copyright © 2010 Laszlo Pandy <laszlok2@gmail.com>
# Copyright © 2011 Michal Hruby <michal.mhr@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import gi
import rb
import time

gi.require_version('Peas', '1.0')
gi.require_version('RB', '3.0')
gi.require_version('Zeitgeist', '2.0')

from gi.repository import GObject, Gio, GLib, Peas, Zeitgeist
from gi.repository import RB

try:
    logger = Zeitgeist.Log.new()

except RuntimeError as e:
    print("Unable to connect to Zeitgeist, won't send events. Reason: '%s'" % e)
    logger = None

class ZeitgeistPlugin(GObject.Object, Peas.Activatable):
    __gtype_name__ = 'ZeitgeistPlugin'
    object = GObject.property(type=GObject.Object)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        print("Loading Zeitgeist plugin...")

        if logger is not None:
            shell = self.object
            shell_player = shell.props.shell_player
            self.__psc_id = shell_player.connect("playing-song-changed", self.playing_song_changed)

            backend_player = shell_player.props.player
            self.__eos_id = backend_player.connect("eos", self.on_backend_eos)

            self.__manual_switch = True
            self.__current_song = None

            event = Zeitgeist.Event.new()
            event.set_property("interpretation", "Source Registration")
            event.set_property("manifestation", Zeitgeist.USER_ACTIVITY)
            event.set_property("actor", "application://org.gnome.Rhythmbox3.desktop")

            datasource = Zeitgeist.DataSource.new()
            datasource.set_unique_id("org.gnome.Rhythmbox3,dataprovider")
            datasource.set_name("Rhythmbox")
            datasource.set_description("Play and organize your music collection")
            datasource.set_event_templates([event,])
            datasource.set_enabled(True)
            datasource.set_timestamp(int(time.time()*1000))

            # Register Rhythmbox as a data source with Zeitgeist
            # engine.
            registry = Zeitgeist.DataSourceRegistry.new()
            registry.register_data_source(datasource)

            # Save references as register_data_source is async.
            self.event = event
            self.datasource = datasource
            self.registry = registry

    @staticmethod
    def get_song_info(db, entry):
        # we don't want the PROP_MEDIA_TYPE, as it doesn't contain mimetype
        # of the audio file itself
        song = {
            "album": entry.get_string(RB.RhythmDBPropType.ALBUM),
            "artist": entry.get_string(RB.RhythmDBPropType.ARTIST),
            "title":  entry.get_string(RB.RhythmDBPropType.TITLE),
            "location": entry.get_playback_uri(),
        }
        return song


    def on_backend_eos(self, backend_player, stream_data, eos_early):
        # EOS signal means that the song changed because the song is over.
        # ie. the user did not explicitly change the song.
        self.__manual_switch = False

    def playing_song_changed(self, shell, entry):
        if self.__current_song is not None:
            self.send_to_zeitgeist_async(self.__current_song, Zeitgeist.LEAVE_EVENT)

        if entry is not None:
            self.send_to_zeitgeist_async(entry, Zeitgeist.ACCESS_EVENT)

        self.__current_song = entry
        GLib.idle_add(self.reset_manual_switch)

    def reset_manual_switch(self):
        """
        After the eos signal has fired, and after the zeitgeist events have
        been sent asynchronously, reset the manual_switch variable.
        """
        self.__manual_switch = True

    def send_to_zeitgeist_async(self, entry, event_type):
        """
        We do async here because the "eos" signal is fired
        *after* the "playing-song-changed" signal.
        We don't know if the song change was manual or automatic
        until we get get the eos signal. If the mainloop goes to
        idle, it means there are no more signals scheduled, so we
        will have already received the eos if it was coming.
        """
        shell = self.object
        db = shell.props.db

        GLib.idle_add(self.send_to_zeitgeist, db, entry, event_type)

    def send_to_zeitgeist(self, db, entry, event_type):
        song = self.get_song_info(db, entry)

        if self.__manual_switch:
            manifest = Zeitgeist.USER_ACTIVITY
        else:
            manifest = Zeitgeist.SCHEDULED_ACTIVITY

        def file_info_complete(obj, res, user_data):
            try:
                fi = obj.query_info_finish(res)
            except:
                return

            uri_mimetype = fi.get_content_type()

            subject = Zeitgeist.Subject.new()
            subject.set_property("uri", song["location"])
            subject.set_property("interpretation", str(Zeitgeist.AUDIO))
            subject.set_property("manifestation", str(Zeitgeist.FILE_DATA_OBJECT))
            subject.set_property("origin", song["location"].rpartition("/")[0])
            subject.set_property("mimetype", uri_mimetype)
            subject.set_property("text", " - ".join([song["title"], song["artist"], song["album"]]))

            event = Zeitgeist.Event.new()
            event.set_property("timestamp", int(time.time()*1000))
            event.set_property("interpretation", str(event_type))
            event.set_property("manifestation", str(manifest))
            event.set_property("actor", "application://org.gnome.Rhythmbox3.desktop")
            event.add_subject(subject)

            logger.insert_event(event)

        f = Gio.file_new_for_uri(song["location"])
        f.query_info_async(Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
                           Gio.FileQueryInfoFlags.NONE,
                           GLib.PRIORITY_DEFAULT,
                           None,
                           file_info_complete,
                           None)

    def do_deactivate(self):
        print("Deactivating Zeitgeist plugin...")
        if logger is not None:
            shell = self.object
            shell_player = shell.props.shell_player
            shell_player.disconnect(self.__psc_id)
            self.__psc_id = None

            backend_player = shell_player.props.player
            backend_player.disconnect(self.__eos_id)
            self.__eos_id = None

            self.__current_song = None



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2010 Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
#

import rb
from gi.repository import Gtk, Gio, GObject, PeasGtk
from gi.repository import RB

import gettext
gettext.install('rhythmbox', RB.locale_dir())

# modes
REPLAYGAIN_MODE_RADIO = 0
REPLAYGAIN_MODE_ALBUM = 1

# number of samples to keep for calculating the average gain
# to apply for tracks that aren't tagged
AVERAGE_GAIN_SAMPLES = 10

class ReplayGainConfig(GObject.Object, PeasGtk.Configurable):
	__gtype_name__ = 'ReplayGainConfig'
	object = GObject.property(type=GObject.Object)

	def do_create_configure_widget(self):
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.replaygain")

		ui_file = rb.find_plugin_file(self, "replaygain-prefs.ui")
		self.builder = Gtk.Builder()
		self.builder.add_from_file(ui_file)

		content = self.builder.get_object("replaygain-prefs")

		combo = self.builder.get_object("replaygainmode")
		combo.props.id_column = 1
		self.settings.bind("mode", combo, "active-id", Gio.SettingsBindFlags.DEFAULT)

		preamp = self.builder.get_object("preamp")
		self.settings.bind("preamp", preamp.props.adjustment, "value", Gio.SettingsBindFlags.GET)
		preamp.connect("value-changed", self.preamp_changed_cb)

		preamp.add_mark(-15.0, Gtk.PositionType.BOTTOM, _("-15.0 dB"))
		preamp.add_mark(0.0, Gtk.PositionType.BOTTOM, _("0.0 dB"))
		preamp.add_mark(15.0, Gtk.PositionType.BOTTOM, _("15.0 dB"))

		limiter = self.builder.get_object("limiter")
		self.settings.bind("limiter", limiter, "active", Gio.SettingsBindFlags.DEFAULT)

		return content

	def preamp_changed_cb(self, preamp):
		RB.settings_delayed_sync(self.settings, self.sync_preamp, preamp)

	def sync_preamp(self, settings, preamp):
		v = preamp.get_value()
		print("preamp gain changed to %f" % v)
		settings['preamp'] = v

GObject.type_register(ReplayGainConfig)
